# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The score function (REINFORCE) gradient estimator of an expectation

from collections.abc import Callable
from functools import partial

import jax
from jax import numpy as jnp

from netket.stats import statistics as mpi_statistics, mean as mpi_mean, Stats
from netket.utils.types import PyTree

from netket.jax import apply_chunked, vjp as nkvjp


def expect(
    log_pdf: Callable[[PyTree, jnp.ndarray], jnp.ndarray],
    expected_fun: Callable[[PyTree, jnp.ndarray], jnp.ndarray],
    pars: PyTree,
    σ: jnp.ndarray,
    *expected_fun_args,
    n_chains: int | None = None,
    chunk_size: int | None = None,
    in_axes: tuple[int | None, ...] | None = None,
) -> tuple[jnp.ndarray, Stats]:
    r"""
    Computes the expectation value over a log-pdf, equivalent to

    .. math::

        \langle f \rangle = \mathbb{E}_{\sigma \sim p(x)}[f(\sigma)] = \sum_{\mathbf{x}} p(\mathbf{x}) f(\mathbf{x})

    where the evaluation of the expectation value is approximated using the sample average, with
    samples :math:`\sigma` that are assumed to be drawn from the probability distribution :math:`p(x)`.

    .. math::

        \langle f \rangle \approx \frac{1}{N} \sum_{i=1}^{N} f(\sigma_i)

    This function ensures that the backward pass is computed correctly, by first differentiating the first equation
    above, and then by approximating the expectation values again using the sample average. The resulting
    backward gradient is

    .. math::

            \nabla \langle f \rangle = \mathbb{E}_{\sigma \sim p(x)}[(\nabla \log p(\sigma)) f(\sigma) + \nabla f(\sigma)]

    where again, the expectation values are comptued using the sample average.

    .. note::

        When using this function together with MPI, you have to pay particular attention. This is because inside the function `f` that is differentiated
        a mean over the MPI ranks (`mpi_mean(term1 + term2, axis=0)`) appears. Therefore, when doing the backward pass this results in a division of the outputs
        from the previous steps by a factor equal to the number of MPI ranks, and so the final gradient on each MPI rank is rescaled as well.
        To cope with this, it is important to sum over the ranks the gradient computed after AD, for example using the function `nk.utils.mpi.mpi_sum_jax`.
        See the following example for more details.

    Example:
        Compute the energy gradient using `nk.jax.expect` on more MPI ranks.

        >>> import netket as nk
        >>> import jax
        >>> import jax.numpy as jnp
        >>>
        >>> hi = nk.hilbert.Spin(s=0.5, N=20)
        >>> graph = nk.graph.Chain(length=20)
        >>> H = nk.operator.IsingJax(hi, graph, h=1.0)
        >>> vstate = nk.vqs.MCState(sampler=nk.sampler.MetropolisLocal(hi, n_chains_per_rank=16), model=nk.models.RBM(alpha=1, param_dtype=complex), n_samples=100000)
        >>>
        >>> afun = vstate._apply_fun
        >>> pars = vstate.parameters
        >>> model_state = vstate.model_state
        >>> log_pdf = lambda params, σ: 2 * afun({"params": params, **model_state}, σ).real
        >>>
        >>> σ = vstate.samples
        >>> σ = σ.reshape(-1, σ.shape[-1])
        >>>
        >>> # The function that we want to differentiate wrt pars and σ
        >>> # Note that we do not want to compute the gradient wrt model_state, so
        >>> # we capture it inside of this function.
        >>> def expect(pars, σ):
        ...
        ...     # The log probability distribution we have generated samples σ from.
        ...     def log_pdf(pars, σ):
        ...         W = {"params": pars, **model_state}
        ...         return 2 * afun(W, σ).real
        ...
        ...     def expected_fun(pars, σ):
        ...         W = {"params": pars, **model_state}
        ...         # Get connected samples
        ...         σp, mels = H.get_conn_padded(σ)
        ...         logpsi_σ = afun(W, σ)
        ...         logpsi_σp = afun(W, σp)
        ...         logHpsi_σ = jax.scipy.special.logsumexp(logpsi_σp, b=mels, axis=1)
        ...         return jnp.exp(logHpsi_σ - logpsi_σ)
        ...     return nk.jax.expect(log_pdf, expected_fun, pars, σ)[0]
        >>>
        >>> E, E_vjp_fun = nk.jax.vjp(expect, pars, σ)
        >>> grad = E_vjp_fun(jnp.ones_like(E))[0]
        >>> grad = jax.tree_util.tree_map(lambda x: nk.utils.mpi.mpi_sum_jax(x)[0], grad)

    Args:
        log_pdf: The log-pdf function from which the samples are drawn. This should output real values, and have a signature
            :code:`log_pdf(pars, σ) -> jnp.ndarray`.
        expected_fun: The function to compute the expectation value of. This should have a signature
            :code:`expected_fun(pars, σ, *expected_fun_args) -> jnp.ndarray`.
        pars: The parameters of the model.
        σ: The samples to compute the expectation value over.
        expected_fun_args: Additional arguments to pass to the expected_fun function (will be differentiated; to avoid
            differentiation, capture them as constants inside of the expected_fun).
        n_chains: The number of chains to use in the computation. If None, the number of chains is inferred from the shape of the input.
        chunk_size: The size of the chunks to use in the computation. If None, no chunking is used.
        in_axes: The axes along which to perform the chunking. If none, only the samples are chunked, otherwise this must be
            the sharding declaration of the samples and the additional arguments to the expected_fun function (must have length
            equal to the number of expected_fun_args + 2).

    Returns:
        A tuple where the first element is the scalar value containing the expectation value, and the second element is
        a :class:`netket.stats.Stats` object containing the statistics (including the mean) of the expectation value.
    """
    return _expect(
        n_chains,
        chunk_size,
        in_axes,
        log_pdf,
        expected_fun,
        pars,
        σ,
        *expected_fun_args,
    )


# log_prob_args and integrand_args are independent of params when taking the
# gradient. They can be continuous or discrete, and they can be pytrees
# Does not support higher-order derivatives yet
@partial(jax.custom_vjp, nondiff_argnums=(0, 1, 2, 3, 4))
def _expect(
    n_chains, chunk_size, in_axes, log_pdf, expected_fun, pars, σ, *expected_fun_args
):
    if chunk_size is not None:
        if in_axes is None:
            # only chunk samples axis
            in_axes = (None, 0) + tuple(None for _ in expected_fun_args)
        else:
            n_args = 2 + len(expected_fun_args)
            assert n_args == len(in_axes)
        expected_fun = apply_chunked(
            expected_fun, chunk_size=chunk_size, in_axes=in_axes
        )

    L_σ = expected_fun(pars, σ, *expected_fun_args)

    if n_chains is not None:
        L_σ = L_σ.reshape((n_chains, -1))

    L̄_σ = mpi_statistics(L_σ)

    return L̄_σ.mean, L̄_σ


def _expect_fwd(
    n_chains, chunk_size, in_axes, log_pdf, expected_fun, pars, σ, *expected_fun_args
):
    if chunk_size is not None:
        if in_axes is None:
            # only chunk samples axis
            in_axes = (None, 0) + tuple(None for _ in expected_fun_args)
        else:
            n_args = 2 + len(expected_fun_args)
            assert n_args == len(in_axes)
        expected_fun = apply_chunked(
            expected_fun, chunk_size=chunk_size, in_axes=in_axes
        )

    L_σ = expected_fun(pars, σ, *expected_fun_args)

    if n_chains is not None:
        L_σ_r = L_σ.reshape((n_chains, -1))
    else:
        L_σ_r = L_σ

    L̄_stat = mpi_statistics(L_σ_r)

    L̄_σ = L̄_stat.mean

    # Use the baseline trick to reduce the variance
    ΔL_σ = L_σ - L̄_σ

    return (L̄_σ, L̄_stat), (pars, σ, expected_fun_args, ΔL_σ)


# TODO: in principle, the gradient of an expectation is another expectation,
# so it should support higher-order derivatives
# But I don't know how to transform log_prob_fun into grad(log_prob_fun) while
# keeping the chunk dimension and without a loop through the chunk dimension
def _expect_bwd(n_chains, chunk_size, in_axes, log_pdf, expected_fun, residuals, dout):
    pars, σ, cost_args, ΔL_σ = residuals
    dL̄, dL̄_stats = dout

    if chunk_size is None:

        def f(pars, σ, *cost_args):
            log_p = log_pdf(pars, σ)
            term1 = jax.vmap(jnp.multiply)(ΔL_σ, log_p)
            term2 = expected_fun(pars, σ, *cost_args)
            out = mpi_mean(term1 + term2, axis=0)
            out = out.sum()
            return out

    else:
        if in_axes is None:
            in_axes = (
                None,
                0,
            ) + tuple(None for _ in cost_args)

        def chunked_f(ΔL_σ, pars, σ, *cost_args):
            log_p = apply_chunked(log_pdf, chunk_size=chunk_size, in_axes=(None, 0))(
                pars, σ
            )
            term1 = jax.vmap(jnp.multiply)(ΔL_σ, log_p)
            term2 = apply_chunked(expected_fun, chunk_size=chunk_size, in_axes=in_axes)(
                pars, σ, *cost_args
            )
            out = mpi_mean(term1 + term2, axis=0)
            out = out.sum()
            return out

        # capture ΔL_σ to not differentiate through it
        f = partial(chunked_f, ΔL_σ)

    _, pb = nkvjp(f, pars, σ, *cost_args)
    grad_f = pb(dL̄)
    return grad_f


_expect.defvjp(_expect_fwd, _expect_bwd)



# Copyright 2021-2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

import numpy as np
from scipy import sparse

import jax.numpy as jnp
from jax.experimental.sparse import JAXSparse, BCOO, BCSR

from netket.operator import AbstractOperator, DiscreteOperator
from netket.utils.optional_deps import import_optional_dependency


class DiscreteJaxOperator(DiscreteOperator):
    r"""Abstract base class for discrete operators that can
    be manipulated inside of jax function transformations.

    Any operator inheriting from this base class follows the
    :class:`netket.operator.DiscreteOperator` interface but
    can additionally be used inside of :func:`jax.jit`,
    :func:`jax.grad`, :func:`jax.vmap` or similar transformations.
    When passed to those functions, jax-compatible operators
    must not be passed as static arguments but as standard
    arguments, and they will not trigger recompilation if
    only the coefficients have changed.

    Most operators can be converted to
    their jax-enabled counterparts by calling the method
    :meth:`~netket.operator.DiscreteJaxOperator.to_jax_operator`.
    Not all operators support this conversion, but as
    :class:`netket.operator.PauliStrings` are flexible, if you
    can convert or write your hamiltonian as a sum of pauli
    strings you will be able to use
    :class:`netket.operator.PauliStringsJax`.


    .. note::

        Jax does not support dynamically varying shapes, so not
        all operators can be written as jax operators, and even
        if they could be written as such, they might generate
        more connected elements than their Numba counterpart.


    .. note::

        :class:`netket.operator.DiscreteJaxOperator` require a
        particular version of the hamiltonian sampling rule,
        :func:`netket.sampler.rules.HamiltonianRuleJax`, that is
        compatible with Jax.


    Defining custom discrete operators that are Jax-compatible
    ----------------------------------------------------------

    This class should be inherited by DiscreteOperators which
    wish to declare jax-compatibility.

    Classes inheriting from `DiscreteJaxOperator`` should be
    declared following a scheme like the following. Do notice
    in particular the declaration of the pytree flattening and
    unflattening, following the standard APIs of Jax discussed
    in the `Jax Pytree documentation <https://jax.readthedocs.io/en/latest/pytrees.html#custom-pytrees-and-initialization>`_.

    .. code-block:: python

         from jax.tree_util import register_pytree_node_class

         @register_pytree_node_class
         class MyJaxOperator(DiscreteJaxOperator):
             def __init__(hilbert, ...):
                 super().__init__(hilbert)

             def tree_flatten(self):
                 array_data = ( ... ) # all arrays
                 struct_data = {'hilbert': self.hilbert,
                                 ... # all constant data
                                 }
                 return array_data, struct_data

             @classmethod
             def tree_unflatten(cls, struct_data, array_data):
                 ...
                 return cls(array_data['hilbert'], ...)

             @property
             def max_conn_size(self) -> int:
                 return ...

             def get_conn_padded(self, x):
                 ...
                 return xp, mels


    """

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def get_conn_padded(self, x: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        r"""Finds the connected elements of the Operator. This method
        can be executed inside of a Jax function transformation.

        Starting from a batch of quantum numbers :math:`x={x_1, ... x_n}` of
        size :math:`B \times M` where :math:`B` size of the batch and :math:`M`
        size of the hilbert space, finds all states :math:`y_i^1, ..., y_i^K`
        connected to every :math:`x_i`.

        Returns a matrix of size :math:`B \times K_{max} \times M` where
        :math:`K_{max}` is the maximum number of connections for every
        :math:`y_i`.

        Args:
            x : A N-tensor of shape :math:`(...,hilbert.size)` containing
                the batch/batches of quantum numbers :math:`x`.

        Returns:
            **(x_primes, mels)**: The connected states x', in a N+1-tensor and an
            N-tensor containing the matrix elements :math:`O(x,x')`
            associated to each x' for every batch.
        """

    def get_conn_flattened(
        self, x: np.ndarray, sections: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        r"""Finds the connected elements of the Operator.

        Starting from a given quantum number :math:`x`, it finds all
        other quantum numbers  :math:`x'` such that the matrix element
        :math:`O(x,x')` is different from zero. In general there will be
        several different connected states :math:`x'` satisfying this
        condition, and they are denoted here :math:`x'(k)`, for
        :math:`k=0,1...N_{\mathrm{connected}}`.

        This is a batched version, where x is a matrix of shape
        :code:`(batch_size,hilbert.size)`.

        Args:
            x: A matrix of shape `(batch_size, hilbert.size)`
                containing the batch of quantum numbers x.
            sections: An array of sections for the flattened x'.
                See numpy.split for the meaning of sections.

        Returns:
            (matrix, array): The connected states x', flattened together in
                a single matrix.
                An array containing the matrix elements :math:`O(x,x')`
                associated to each x'.

        """
        xp, mels = self.get_conn_padded(x)
        n_conns = mels.shape[1]
        xp = xp.reshape(-1, xp.shape[-1])
        mels = mels.reshape(-1)
        sections[:] = n_conns
        return xp, mels

    def n_conn(self, x, out=None) -> np.ndarray:
        r"""Return the number of states connected to x.

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            out (array): If None an output array is allocated.

        Returns:
            array: The number of connected states x' for each x[i].

        """
        if out is None:
            out = jnp.full(x.shape[0], self.max_conn_size, dtype=np.int32)
        else:
            out[:] = self.max_conn_size
        return out

    def to_sparse(self) -> JAXSparse:
        r"""Returns the sparse matrix representation of the operator. Note that,
        in general, the size of the matrix is exponential in the number of quantum
        numbers, and this operation should thus only be performed for
        low-dimensional Hilbert spaces or sufficiently sparse operators.

        This method requires an indexable Hilbert space.

        Returns:
            The sparse jax matrix representation of the operator.
        """

        # TODO: If the operator get_conn_padded uses shard_map, the
        # replication of all_states will lead to a crash when
        # the n_samples cannot be divided by the number of ranks.
        # this should be fixed.
        x = self.hilbert.all_states()
        n = x.shape[0]
        xp, mels = self.get_conn_padded(x)
        a = mels.ravel()
        i = np.broadcast_to(np.arange(n)[..., None], mels.shape).ravel()
        j = self.hilbert.states_to_numbers(xp).ravel()
        ij = np.concatenate((i[:, None], j[:, None]), axis=1)
        return BCSR.from_bcoo(BCOO((a, ij), shape=(n, n)))

    def to_dense(self) -> np.ndarray:
        r"""Returns the dense matrix representation of the operator. Note that,
        in general, the size of the matrix is exponential in the number of quantum
        numbers, and this operation should thus only be performed for
        low-dimensional Hilbert spaces or sufficiently sparse operators.

        This method requires an indexable Hilbert space.

        Returns:
            The dense matrix representation of the operator as a jax Array.
        """
        return self.to_sparse().todense()

    def to_qobj(self):  # -> "qutip.Qobj"
        r"""Convert the operator to a qutip's Qobj.

        Returns:
            A :class:`qutip.Qobj` object.
        """
        qutip = import_optional_dependency("qutip", descr="to_qobj")

        # QuTiP does not like jax sparse matrices, so we convert to scipy sparse
        # by hand
        sparse_mat_jax = self.to_sparse()
        sparse_mat_scipy = sparse.csr_matrix(
            (
                sparse_mat_jax.data,
                sparse_mat_jax.indices,
                sparse_mat_jax.indptr,
            ),
            shape=sparse_mat_jax.shape,
        )
        return qutip.Qobj(
            sparse_mat_scipy, dims=[list(self.hilbert.shape), list(self.hilbert.shape)]
        )

    def __matmul__(self, other):
        if isinstance(other, JAXSparse):
            return self.apply(other.todense())
        elif isinstance(other, np.ndarray) or isinstance(other, jnp.ndarray):
            return self.apply(other)
        elif isinstance(other, AbstractOperator):
            return self._op__matmul__(other)
        else:
            return NotImplemented

    def __rmatmul__(self, other):
        if (
            isinstance(other, np.ndarray)
            or isinstance(other, jnp.ndarray)
            or isinstance(other, JAXSparse)
        ):
            return NotImplemented
        elif isinstance(other, AbstractOperator):
            return self._op__rmatmul__(other)
        else:
            return NotImplemented

    def to_jax_operator(self) -> "DiscreteJaxOperator":
        """
        Return the JAX version of this operator.

        If this is a JAX operator does nothing.
        """
        return self



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import (
    VariationalState,
    VariationalMixedState,
    expect,
    expect_and_grad,
    expect_and_forces,
)

from .mc import MCState, MCMixedState, get_local_kernel_arguments, get_local_kernel
from .full_summ import FullSumState

_deprecations = {
    # May 2023
    "ExactState": (
        "netket.vqs.ExactState is deprecated: use netket.vqs.FullSumState (netket >= 3.12)",
        FullSumState,
    ),
}

from netket.utils import _hide_submodules
from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr

_hide_submodules(__name__, ignore=["experimental"], hide_folder=["mc"])
__getattr__ = _deprecation_getattr(__name__, _deprecations)
del _deprecation_getattr



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
from jax import numpy as jnp

from netket.utils.types import PyTree
from netket.errors import RealQGTComplexDomainError
from netket.jax._utils_tree import RealImagTuple


def check_valid_vector_type(x: PyTree, target: PyTree):
    """
    Raises a TypeError if x is complex where target is real, because it is not
    supported by QGTOnTheFly and the imaginary part would be discarded after
    anyhow.
    """

    def check(x, target, target_im=None):
        par_iscomplex = jnp.iscomplexobj(x)

        # Account for split real-imaginary part in Jacobian*** methods
        vec_iscomplex = target_im is not None or jnp.iscomplexobj(target)

        if not par_iscomplex and vec_iscomplex:
            raise RealQGTComplexDomainError()

    try:
        if isinstance(target, RealImagTuple):
            jax.tree_util.tree_map(check, x, target.real, target.imag)
        else:
            jax.tree_util.tree_map(check, x, target)
    except ValueError:
        # catches jax tree map errors
        pars_struct = jax.tree_util.tree_map(
            lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), x
        )
        vec_struct = jax.tree_util.tree_map(
            lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), target
        )

        raise ValueError(
            "PyTree mismatch: Parameters have shape \n\n"
            f"{pars_struct}\n\n"
            "but the vector has shape \n\n"
            f"{vec_struct}\n\n"
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any
from collections.abc import Callable
import warnings

import jax
from jax import numpy as jnp
from flax.core.scope import CollectionFilter, DenyList  # noqa: F401

from netket import jax as nkjax
from netket.operator import AbstractOperator
from netket.operator._abstract_observable import AbstractObservable
from netket.stats import Stats, statistics
from netket.utils import mpi
from netket.utils.types import PyTree

from netket.vqs import expect_and_forces
from netket.vqs.mc import (
    get_local_kernel,
    get_local_kernel_arguments,
)

from .state import MCState


# If batch_size is unspecified, set it to None
@expect_and_forces.dispatch
def expect_and_forces_chunking_unspecified(  # noqa: F811
    vstate: MCState,
    operator: AbstractObservable,
    **kwargs,
):
    return expect_and_forces(vstate, operator, None, **kwargs)


# if no implementation exists for batched, run the code unbatched
@expect_and_forces.dispatch(precedence=-10)
def expect_and_forces_fallback(  # noqa: F811
    vstate: MCState,
    operator: AbstractObservable,
    chunk_size: Any,
    *args,
    **kwargs,
):
    warnings.warn(
        f"Ignoring chunk_size={chunk_size} for expect_and_forces method with signature "
        f"({type(vstate)}, {type(operator)}) because no implementation supporting "
        f"chunking for this signature exists.",
        stacklevel=2,
    )

    return expect_and_forces(vstate, operator, None, *args, **kwargs)


@expect_and_forces.dispatch
def expect_and_forces_impl(  # noqa: F811
    vstate: MCState,
    Ô: AbstractOperator,
    chunk_size: int,
    *,
    mutable: CollectionFilter = False,
) -> tuple[Stats, PyTree]:
    σ, args = get_local_kernel_arguments(vstate, Ô)

    local_estimator_fun = get_local_kernel(vstate, Ô, chunk_size)

    Ō, Ō_grad, new_model_state = forces_expect_hermitian_chunked(
        chunk_size,
        local_estimator_fun,
        vstate._apply_fun,
        mutable,
        vstate.parameters,
        vstate.model_state,
        σ,
        args,
    )

    if mutable is not False:
        vstate.model_state = new_model_state

    return Ō, Ō_grad


@partial(jax.jit, static_argnums=(0, 1, 2, 3))
def forces_expect_hermitian_chunked(
    chunk_size: int,
    local_value_kernel_chunked: Callable,
    model_apply_fun: Callable,
    mutable: CollectionFilter,
    parameters: PyTree,
    model_state: PyTree,
    σ: jnp.ndarray,
    local_value_args: PyTree,
) -> tuple[PyTree, PyTree]:
    σ_shape = σ.shape
    if jnp.ndim(σ) != 2:
        σ = σ.reshape((-1, σ_shape[-1]))

    n_samples = σ.shape[0] * mpi.n_nodes

    O_loc = local_value_kernel_chunked(
        model_apply_fun,
        {"params": parameters, **model_state},
        σ,
        local_value_args,
        chunk_size=chunk_size,
    )

    Ō = statistics(O_loc.reshape(σ_shape[:-1]).T)

    O_loc -= Ō.mean

    # Then compute the vjp.
    # Code is a bit more complex than a standard one because we support
    # mutable state (if it's there)
    if mutable is False:
        vjp_fun_chunked = nkjax.vjp_chunked(
            lambda w, σ: model_apply_fun({"params": w, **model_state}, σ),
            parameters,
            σ,
            conjugate=True,
            chunk_size=chunk_size,
            chunk_argnums=1,
            nondiff_argnums=1,
        )
        new_model_state = None
    else:
        raise NotImplementedError

    Ō_grad = vjp_fun_chunked(
        (jnp.conjugate(O_loc) / n_samples),
    )[0]

    Ō_grad, _ = mpi.mpi_sum_jax(Ō_grad)

    return Ō, Ō_grad, new_model_state



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Internal utility functions to support jax sharding natively within netket.
All functions in here are not part of the public API, internal, and may change without warning.
"""

import math
from functools import partial, wraps
import contextlib


import jax
import jax.numpy as jnp
from jax.tree_util import Partial
from jax.sharding import (
    Mesh,
    PartitionSpec as P,
    PositionalSharding,
)
from jax.experimental.shard_map import shard_map
from jax.util import safe_zip

from netket.utils import config, mpi
from netket.utils.deprecation import warn_deprecation


_identity = lambda x: x


def _prepare_mask(n, n_pad):
    return jnp.ones(n + n_pad, dtype=bool).at[-n_pad:].set(0)


def distribute_to_devices_along_axis(
    inp_data, axis=0, pad: bool = False, pad_value=None, devices=None
):
    """
    Distribute a local array equally along an axis to multiple jax devices devices

     .. note:
        Does nothing if netket.config.netket_experimental_sharding=False.

     .. note:
        Each jax process needs to have the whole array (parts not belonging to it can be filled with garbage).

    Args:
        inp_data: the full array (on every process)
        axis: (optional) axis along which to distribute
        pad: If True: pad the input data along axis to the next multiple of the number of devices
              If False (default): no padding; the size of the axis in inp_data needs to be divisible by the number of devices.
        pad_value: value to pad with (optional, only used if pad=True)
        devices: (optional) list of jax devices. Defaults to all available devices

    Returns:
        out_data: a distributed jax.Array
        mask: a mask indicating wether a given element is part of the original data (True) of of the padding (False)
              only returned if pad=True
    """
    if devices is None:
        devices = jax.devices()

    if config.netket_experimental_sharding:
        if pad:
            n = inp_data.shape[0]
            # pad to the next multiple of device_count
            device_count = jax.device_count()
            n_pad = math.ceil(inp_data.shape[0] / device_count) * device_count - n
            inp_data = jnp.pad(inp_data, ((0, n_pad), (0, 0)))
            if pad_value is not None and n_pad > 0:
                inp_data = inp_data.at[-n_pad:].set(pad_value)

        shape = [
            1,
        ] * inp_data.ndim
        shape[axis] = -1
        sharding = PositionalSharding(devices).reshape(shape)
        out_data = jax.jit(_identity, out_shardings=sharding)(inp_data)
        # TODO support gspmdsharding in numba wrapper and use this
        # out_data = jax.jit(jax.lax.with_sharding_constraint, static_argnums=1)(
        #     inp_data, sharding
        # )

        if pad:
            if n_pad > 0:  # type: ignore
                mask = jax.jit(
                    _prepare_mask,
                    out_shardings=sharding.reshape(-1),
                    static_argnums=(0, 1),
                )(
                    n, n_pad
                )  # type: ignore
            else:
                mask = None
            return out_data, mask
        else:
            return out_data
    else:
        return inp_data


# TODO consider merging this with distribute_to_devices_along_axis
@partial(jax.jit, static_argnames="axis")
def shard_along_axis(x, axis: int):
    """
    When running with experimental sharding mode, calls
    :func:`jax.lax.with_sharding_constraint` with a
    :class:`jax.sharding.PositionalSharding` sharded along the given axis.

    Args:
        x: An array
        axis: the axis to be sharded
    """
    if config.netket_experimental_sharding and jax.device_count() > 1:
        # Shard shape is (1, 1, 1, -1, 1, 1) where -1 is the axis
        shard_shape = [1 for _ in range(x.ndim)]
        shard_shape[axis] = -1

        x = jax.lax.with_sharding_constraint(
            x, PositionalSharding(jax.devices()).reshape(tuple(shard_shape))
        )
    return x


@jax.jit
def with_samples_sharding_constraint(x, shape=None):
    """
    ensure the input x is sharded along axis 0 on all devices
    works both outside and inside of jit
    """
    warn_deprecation(
        "with_samples_sharding_constraint is deprecated in favour of nk.jax.sharding.shard_along_axis(x, axis=0)"
    )

    return shard_along_axis(x, 0)


def extract_replicated(t):
    """
    Extract the value of a fully replicated global device array.

    Args:
        t: a jax Array (or a pytree of jax Arrays)

    Returns:
        A locally adressable representation of t
    """

    def _extract_replicated(x):
        if isinstance(x, jax.Array) and not x.is_fully_addressable:
            if not x.is_fully_replicated:
                raise RuntimeError(
                    "Expected a fully replicated array, but found one that is not. You should gather the array first. If you are not developing custom logic, please open a bug report with a reproducer."
                )
            return x.addressable_data(0)
        else:
            return x

    return jax.tree_util.tree_map(_extract_replicated, t)


def gather(x):
    """
    Make a sharded array fully replicated by gathering all parts on every device.

    Args:
        x: potentially unreplicated jax.Array with PositionalSharding

    Returns:
        fully replicated array
    """

    if not isinstance(x, jax.Array):
        # TODO in the future we could chagne it to just return x unchanged
        # but for now we error if x is not a jax array to ensure gather is used correctly
        raise RuntimeError("gather can only be applied to a jax.Array")
    elif x.is_fully_replicated:  # includes SingleDeviceSharding
        return x
    elif isinstance(x.sharding, jax.sharding.GSPMDSharding):
        # x.sharding.device_set has arbitrary order
        # Hardcode all devices until I figure out a way to deduce the order from x
        out_shardings = (
            PositionalSharding(jax.devices()).replicate().reshape((1,) * x.ndim)
        )
        # TODO support gspmdsharding in numba wrapper and use this
        # out_shardings = x.sharding.get_replicated(jax.devices())
    elif isinstance(x.sharding, PositionalSharding):
        out_shardings = x.sharding.replicate()
    else:
        raise NotImplementedError(
            "Gather is only compatible with PositionalSharding and GSPMDSharding,"
            f" but array has {x.sharding}. Please open a feature request."
        )
    return jax.jit(_identity, out_shardings=out_shardings)(x)
    # TODO support gspmdsharding in numba wrapper and use this
    # return jax.jit(jax.lax.with_sharding_constraint, static_argnums=1)(x, out_shardings)


SHARD_MAP_STACK_LEVEL: int = 0
"""
A counter used to keep track of how many levels deep we are in shard_map.

This should not be modified directly, but controlled through the context manager
_increase_SHARD_MAP_STACK_LEVEL.
"""


@contextlib.contextmanager
def _increase_SHARD_MAP_STACK_LEVEL():
    """
    A context manager used by `sharding_decorator` to keep track of how many nested
    sharded function calls we are in.

    In pratice, this is used to ensure that only the outermost function is sharded, and following
    calls are not sharded.
    """
    global SHARD_MAP_STACK_LEVEL
    try:
        SHARD_MAP_STACK_LEVEL += 1
        yield
    except Exception as e:
        raise e
    finally:
        SHARD_MAP_STACK_LEVEL -= 1


def sharding_decorator(f, sharded_args_tree, reduction_op_tree=False, **kwargs):
    """
    A decorator which wraps a function so that it is evaluated on every shard of the distributed arguments,
    and the output is either returned sharded, or can be reduced with a collective operation.

    This is essentially a fancy wrapper around jax.experimental.shard_map,
    meant to be used to wrap the `chunked` parts of netket (vmap_chunked, vjp_chunked, ...), so that the
    computations are computed in chunks on every devices shard (and not in chunks of the whole array).

    .. warning::
        Intended for netket internal use only, the interface might change in the future based on our requirements.

    .. note:
        if `netket.config.netket_experimental_sharding=False` it returns the unchanged original function

    .. note:
        If nested functions are decorated with this decorator, only the outermost one is sharded and the internal
        ones are ignored.

    Args:
        f: a function
        sharded_args_tree: a tuple / tuple of pyrtrees of length of the number of args in f
            containing True/False indicating that each input in the argumens of f is:
                True: sharded on axis 0 (True)
                False: assumed to be replicated
                'key': A single jax.random.key. It is split across devices so that the function executed on every device is passed a different key.
                       If the argument is already a sharded array of keys use True instead.
            the args of f are flattened according to sharded_args_tree, so if an arg is a pytree a single True/False is assumed the whole tree
        reduction_op_tree: a tuple/pyrtree of reduction_op, where for each output:
            reduction_op is e.g. jax.lax.psum if it is to be reduced, then f_wrapped returns a replicated array
            reduction op is False if it is not to be reduced, then f_wrapped returns a sharded array
            reduction op is True if it is not an array/pytree, then it is returned as python object

    Returns :
        f_wrapped: wrapped version of f


    Example:


        %env NETKET_EXPERIMENTAL_SHARDING=1

        import jax
        import jax.numpy as jnp
        from jax.sharding import Mesh, PartitionSpec as P
        from jax.experimental.shard_map import shard_map
        from jax.tree_util import Partial
        from functools import partial
        from netket import config
        from netket.jax.sharding import sharding_decorator

        assert config.netket_experimental_sharding is True
        assert jax.device_count() > 1

        def expensive_elementwise_function(x, c):
            return x + c

        def looped_computation(x, c=1, f=expensive_elementwise_function):
            y = jnp.zeros_like(x)
            for i in range(len(x)):
                y = y.at[i].set(f(x[i], c))
            return y

        x = jax.jit(jnp.ones, out_shardings=jax.sharding.PositionalSharding(jax.devices()), static_argnums=0)(jax.device_count()*5)
        c = jax.jit(jnp.ones, out_shardings=jax.sharding.PositionalSharding(jax.devices()).replicate(), static_argnums=0)(())

        # if we were to run `looped_computation(x)`` with the sharded x, it would formally be computed sequentially for all elements device per device,
        # if we  jit, i.e. `jax.jit(looped_computation)(x)`` the output sharding would just be replicated, jax just computes everything replicated on every device.
        # However we want to compute sequentially on every device in parallel

        # one way to do this is by reshaping with the number of devices, and moving the axes (and in general might require changing the function)
        x_per_device = x.reshape(len(x.devices()), -1)
        x_per_device = jnp.swapaxes(x_per_device, 0,1) # make the devices the second axis
        y_per_device = jax.jit(jax.vmap(looped_computation, in_axes=1, out_axes=1))(x_per_device)
        y_per_device = jnp.swapaxes(y_per_device, 0,1)  # output swap back
        y_flat = y_per_device.ravel()
        jax.debug.visualize_array_sharding(y_flat)

        # we can achieve the same in 1 line with sharding_decorator:
        y = jax.jit(sharding_decorator(looped_computation, sharded_args_tree=(True,)))(x)
        jax.debug.visualize_array_sharding(y)

        # it also supports a mixture of arrays with PositionalSharding along an axis and replicated sharding:
        y = jax.jit(sharding_decorator(looped_computation, sharded_args_tree=(True, False)))(x, c)
        jax.debug.visualize_array_sharding(y)

        #
        # furthermore sharding decorator supports reduction operations on the output:
        #
        def looped_computation2(x):
            y = looped_computation(x)
            return y.sum(axis=0)

        # again the manual version:
        y2_per_device = jax.jit(jax.vmap(looped_computation2, in_axes=1, out_axes=0))(x_per_device)
        # take sum by hand over devices
        y2_flat = y2_per_device.sum(axis=0)

        # and with sharding_decorator:
        y2 = jax.jit(sharding_decorator(looped_computation2, sharded_args_tree=(True,), reduction_op_tree=jax.lax.psum))(x)

        # which internally wraps the function like this:
        mesh = Mesh(jax.devices(), axis_names=("i"))
        in_specs = P("i")
        out_specs = P()
        @partial(shard_map, mesh=mesh, in_specs=in_specs, out_specs=out_specs)
        def _f(x):
            res = looped_computation2(x)
            res = jax.lax.psum(res, axis_name="i")
            return res

        #
        # Furthermore it supports non-array outputs with the special choice of reduction_op True:
        #
        def looped_computation3(x):
            y = looped_computation(x)
            some_python_object = {1,2,3}
            return y, some_python_object
        # here we cannot jit, as the static some_python_object cannot be returned from a jitted function
        # this is meant for calling inside of another jitted function, where e.g. some computation depends on the object
        # but is not retuned at the end (or is with the same trick we use here)
        y3, my_python_object = sharding_decorator(looped_computation3, sharded_args_tree=(True,), reduction_op_tree=(False, True))(x)

        # Internally it does something like this:

        mesh = Mesh(jax.devices(), axis_names=("i"))
        in_specs = P("i")
        out_specs = P('i'), P()
        @partial(shard_map, mesh=mesh, in_specs=in_specs, out_specs=out_specs)
        def _f(x):
            some_python_object = {1,2,3}
            return x, Partial(partial(lambda x: x, some_python_object))
        y, obj_wrapped = _f(x)
        obj = obj_wrapped()

        # Finally we note that for non-array inputs, which are not yet supported by the jax experimental shard_map we
        # (automatically) wrap them in the metadata of a Partial if they don't have a dtype attribute

        # In the following, for the sake of better understanding we describe how we would have to do this manually

        # If it is a function we could just wrap it in a Partial directly:
        # NB: for the Partial ith no .args it is irrelevant if we use True or False in sharded_args_tree, jax just sees it as an empty array

        y = sharding_decorator(looped_computation, sharded_args_tree=(True, False, False))(x, c, Partial(expensive_elementwise_function))

        # In general we hide it in a partial inside of a partial which we have to call inside of the function
        # the same trick as for reduction op True above

        # here we use Partial(partial(lambda x: x, obj)), iirc just a Partial(lambda x: obj)) didn't work in some cases

        def looped_computation4(x, obj_wrapped):
            obj = obj_wrapped()
            if obj:
                y = looped_computation(x)
            else:
                y = x
            return y
        obj = True
        obj_wrapped = Partial(partial(lambda x: x, obj))
        y4 = jax.jit(sharding_decorator(looped_computation4, sharded_args_tree=(True, False)))(x, obj_wrapped)
    """
    if config.netket_experimental_sharding:  # type: ignore
        if not isinstance(sharded_args_tree, tuple):
            sharded_args_tree = (sharded_args_tree,)
        sharded_args, args_treedef = jax.tree_util.tree_flatten(sharded_args_tree)
        reduction_op, out_treedef = jax.tree_util.tree_flatten(reduction_op_tree)

        @wraps(f)
        def _fun(*args_orig):
            global SHARD_MAP_STACK_LEVEL
            # Jax 0.4.28 does not support nested shard_map calls, so we bail out eaerly
            # if we are already inside of a shard map call
            if SHARD_MAP_STACK_LEVEL > 0:
                return f(*args_orig)

            args = args_treedef.flatten_up_to(args_orig)

            _sele = lambda cond, xs: tuple(x for c, x in safe_zip(cond, xs) if c)
            _not = lambda t: tuple(not x for x in t)
            _sele2 = lambda cond, x, y: tuple(x if c else y for c in cond)

            # PRNGKey treatment 1/2
            args = tuple(
                jax.random.split(a, jax.device_count()) if c == "key" else a
                for a, c in safe_zip(args, sharded_args)
            )

            # workaround for shard_map not supporting non-array args part 1/2
            nonarray_args = tuple(not hasattr(a, "dtype") for a in args)
            args = tuple(
                Partial(partial(lambda x: x, a)) if c else a
                for a, c in safe_zip(args, nonarray_args)
            )

            mesh = Mesh(jax.devices(), axis_names=("i"))
            in_specs = _sele2(sharded_args, P("i"), P())
            out_specs = out_treedef.unflatten(_sele2(reduction_op, P(), P("i")))

            @partial(
                shard_map, mesh=mesh, in_specs=in_specs, out_specs=out_specs, **kwargs
            )
            def _f(*args):
                # workaround for shard_map not supporting non-array args part 2/2
                args = tuple(a() if c else a for a, c in safe_zip(args, nonarray_args))

                # PRNGKey treatment 2/2
                args = tuple(
                    a[0] if c == "key" else a for a, c in safe_zip(args, sharded_args)
                )

                res = f(*args_treedef.unflatten(args))

                # apply reductions
                # _id = lambda x: x
                # _wrap = lambda x: Partial(lambda : x)
                def _sele_op(o):
                    if o is False:
                        return lambda x: x
                    if o is True:
                        return lambda x: Partial(partial(lambda x: x, x))
                    else:
                        return partial(
                            jax.tree_util.tree_map, partial(o, axis_name="i")
                        )

                reductions = [_sele_op(o) for o in reduction_op]
                res = out_treedef.flatten_up_to(res)
                res = [f(r) for f, r in safe_zip(reductions, res)]
                res = out_treedef.unflatten(res)
                return res

            # We are sure that, so far, we have SHARD_MAP_STACK_LEVEL=0, so we can
            # safely call a @shard_map decorated function. The context manager
            # makes sure we only shard the outermost call.
            with _increase_SHARD_MAP_STACK_LEVEL():
                res = _f(*args)

            res = out_treedef.flatten_up_to(res)
            res = [a() if c is True else a for a, c in safe_zip(res, reduction_op)]
            res = out_treedef.unflatten(res)
            return res

        return _fun

    return f


def device_count_per_rank() -> int:
    """
    Helper functions which returns the number of jax devices netket will use for every
    MPI rank.

    Returns:
        jax.device_count() if config.netket_experimental_sharding is True, and 1 otherwise
    """
    if config.netket_experimental_sharding:  # type: ignore
        if mpi.n_nodes > 1:
            # this should never be triggered as we disable mpi when sharding
            raise NotImplementedError("hybrid mpi and sharding is not not supported")
        return jax.device_count()
    else:  # mpi or serial
        return 1


def device_count() -> int:
    """
    Helper functions which returns the TOTAL number of jax devices netket will use.

    Returns:
        jax.device_count() if config.netket_experimental_sharding is True, and mpi.rank otherwise.
    """
    return mpi.n_nodes * device_count_per_rank()



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import numpy as np
from flax import linen as nn
from flax.linen.dtypes import promote_dtype

import jax
from jax import lax
from jax import numpy as jnp
from jax.nn.initializers import lecun_normal, zeros

from netket.utils.types import Array, DType, NNInitFunc

default_kernel_init = lecun_normal()


def wrap_kernel_init(kernel_init, mask):
    """Correction to LeCun normal init."""

    corr = jnp.sqrt(mask.size / mask.sum())

    def wrapped_kernel_init(*args):
        return corr * mask * kernel_init(*args)

    return wrapped_kernel_init


# This is copy-pasted from flax.linen.linear in order to vendor it
def _conv_dimension_numbers(input_shape):
    """Computes the dimension numbers based on the input shape."""
    ndim = len(input_shape)
    lhs_spec = (0, ndim - 1, *tuple(range(1, ndim - 1)))
    rhs_spec = (ndim - 1, ndim - 2, *tuple(range(0, ndim - 2)))
    out_spec = lhs_spec
    return lax.ConvDimensionNumbers(lhs_spec, rhs_spec, out_spec)


class MaskedDense1D(nn.Module):
    """1D linear transformation module with mask for autoregressive NN."""

    features: int
    """output feature density, should be the last dimension."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the weight matrix."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    @nn.compact
    def __call__(self, inputs: jax.Array) -> Array:
        """
        Applies a masked linear transformation to the inputs.

        Args:
          inputs: input data with dimensions (batch, length, features).

        Returns:
          The transformed data.
        """
        if inputs.ndim == 2:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)
        else:
            is_single_input = False

        batch, size, in_features = inputs.shape
        inputs = inputs.reshape((batch, size * in_features))

        if self.use_bias:
            bias = self.param(
                "bias", self.bias_init, (size, self.features), self.param_dtype
            )
        else:
            bias = None

        mask = np.ones((size, size), dtype=self.param_dtype)
        mask = np.triu(mask, self.exclusive)
        mask = np.kron(
            mask, np.ones((in_features, self.features), dtype=self.param_dtype)
        )
        mask = jnp.asarray(mask)  # type: ignore[no-redef]

        kernel = self.param(
            "kernel",
            wrap_kernel_init(self.kernel_init, mask),
            (size * in_features, size * self.features),
            self.param_dtype,
        )

        inputs, mask, kernel, bias = promote_dtype(
            inputs, mask, kernel, bias, dtype=None
        )

        y = lax.dot(inputs, mask * kernel, precision=self.precision)

        y = y.reshape((batch, size, self.features))

        if is_single_input:
            y = y.squeeze(axis=0)

        if self.use_bias:
            y = y + bias

        return y


class MaskedConv1D(nn.Module):
    """1D convolution module with mask for autoregressive NN."""

    features: int
    """number of convolution filters."""
    kernel_size: int
    """length of the convolutional kernel."""
    kernel_dilation: int
    """dilation factor of the convolution kernel."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    feature_group_count: int = 1
    """if specified, divides the input features into groups (default: 1)."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the convolutional kernel."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    @nn.compact
    def __call__(self, inputs: Array) -> Array:
        """
        Applies a masked convolution to the inputs.
        For 1D convolution, there is not really a mask. We only need to apply
        appropriate padding.

        Args:
          inputs: input data with dimensions (batch, length, features).

        Returns:
          The convolved data.
        """
        kernel_size = self.kernel_size - self.exclusive
        dilation = self.kernel_dilation

        if inputs.ndim == 2:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)
        else:
            is_single_input = False

        in_features = inputs.shape[-1]
        assert in_features % self.feature_group_count == 0
        kernel_shape = (
            kernel_size,
            in_features // self.feature_group_count,
            self.features,
        )

        if self.use_bias:
            bias = self.param(
                "bias", self.bias_init, (self.features,), self.param_dtype
            )
        else:
            bias = None

        kernel = self.param("kernel", self.kernel_init, kernel_shape, self.param_dtype)

        inputs, kernel, bias = promote_dtype(inputs, kernel, bias, dtype=None)

        if self.exclusive:
            inputs = inputs[:, :-dilation, :]

        # Zero padding
        y = jnp.pad(
            inputs,
            (
                (0, 0),
                ((kernel_size - (not self.exclusive)) * dilation, 0),
                (0, 0),
            ),
        )

        dimension_numbers = _conv_dimension_numbers(inputs.shape)
        y = lax.conv_general_dilated(
            y,
            kernel,  # type: ignore[arg-type]
            window_strides=(1,),
            padding="VALID",
            lhs_dilation=(1,),
            rhs_dilation=(dilation,),
            dimension_numbers=dimension_numbers,
            feature_group_count=self.feature_group_count,
            precision=self.precision,
        )

        if is_single_input:
            y = y.squeeze(axis=0)

        if self.use_bias:
            y = y + bias

        return y


class MaskedConv2D(nn.Module):
    """2D convolution module with mask for autoregressive NN."""

    features: int
    """number of convolution filters."""
    kernel_size: tuple[int, int]
    """shape of the convolutional kernel `(h, w)`. Typically, `h = w // 2 + 1`."""
    kernel_dilation: tuple[int, int]
    """a sequence of 2 integers, giving the dilation factor to
    apply in each spatial dimension of the convolution kernel."""
    exclusive: bool
    """True if an output element does not depend on the input element at the same index."""
    feature_group_count: int = 1
    """if specified, divides the input features into groups (default: 1)."""
    use_bias: bool = True
    """whether to add a bias to the output (default: True)."""
    param_dtype: DType = jnp.float64
    """the dtype of the computation (default: float64)."""
    precision: Any = None
    """numerical precision of the computation, see :class:`jax.lax.Precision` for details."""
    kernel_init: NNInitFunc = default_kernel_init
    """initializer for the convolutional kernel."""
    bias_init: NNInitFunc = zeros
    """initializer for the bias."""

    def setup(self):
        kernel_h, kernel_w = self.kernel_size
        mask = np.ones((kernel_h, kernel_w, 1, 1), dtype=self.param_dtype)
        mask[-1, kernel_w // 2 + (not self.exclusive) :] = 0
        self.mask = jnp.asarray(mask)

    @nn.compact
    def __call__(self, inputs: Array) -> Array:
        """
        Applies a masked convolution to the inputs.

        Args:
          inputs: input data with dimensions (batch, width, height, features).

        Returns:
          The convolved data.
        """
        kernel_h, kernel_w = self.kernel_size
        dilation_h, dilation_w = self.kernel_dilation
        ones = (1, 1)

        if inputs.ndim == 3:
            is_single_input = True
            inputs = jnp.expand_dims(inputs, axis=0)
        else:
            is_single_input = False

        in_features = inputs.shape[-1]
        assert in_features % self.feature_group_count == 0
        kernel_shape = self.kernel_size + (
            in_features // self.feature_group_count,
            self.features,
        )

        mask = self.mask

        kernel = self.param(
            "kernel",
            wrap_kernel_init(self.kernel_init, mask),
            kernel_shape,
            self.param_dtype,
        )

        if self.use_bias:
            bias = self.param(
                "bias", self.bias_init, (self.features,), self.param_dtype
            )
        else:
            bias = None

        inputs, mask, kernel, bias = promote_dtype(
            inputs, mask, kernel, bias, dtype=None
        )

        # Zero padding
        y = jnp.pad(
            inputs,
            (
                (0, 0),
                ((kernel_h - 1) * dilation_h, 0),
                (kernel_w // 2 * dilation_w, (kernel_w - 1) // 2 * dilation_w),
                (0, 0),
            ),
        )

        dimension_numbers = _conv_dimension_numbers(inputs.shape)
        y = lax.conv_general_dilated(
            y,
            mask * kernel,
            window_strides=ones,
            padding="VALID",
            lhs_dilation=ones,
            rhs_dilation=self.kernel_dilation,
            dimension_numbers=dimension_numbers,
            feature_group_count=self.feature_group_count,
            precision=self.precision,
        )

        if is_single_input:
            y = y.squeeze(axis=0)

        if self.use_bias:
            y = y + bias

        return y



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tools to compare and hash floating point numbers safely.
"""

import numpy as np
from netket.utils.types import Array, Union


def comparable(
    x: Array, *, bin_density: int = 3326400, offset: float = 5413 / 15629
) -> Array:
    """
    Casts a floating point input to integer-indexed bins that are safe to compare
    or hash.

    Arguments:
        x: the float array to be converted
        bin_density: the inverse width of each bin. When binning rational numbers,
            it's best to use a multiple of all expected denominators `bin_density`.
            The default is :math:`3326400 = 2^6\times 3^3\times 5^2\times 7\times 11`.
        offset: constant offset added to `bin_density * x` before rounding. To minimise
            the chances of "interesting" numbers appearing on bin boundaries, it's
            best to use a rational number with a large prime denominator.
            The default is 5413/15629, both are primes.

    Returns:
        `x * bin_density + offset` rounded to an integer

    Example:

        >>> comparable([0.0, 0.3, 0.30000001, 1.3])
        array([      0,  997920,  997920, 4324320])
    """
    return np.asarray(np.rint(np.asarray(x) * bin_density + offset), dtype=int)


def comparable_periodic(
    x: Array,
    where: Union[Array, bool] = True,
    *,
    bin_density: int = 3326400,
    offset: float = 5413 / 15629,
) -> Array:
    """
    Casts the fractional parts of floating point input to integer-indexed bins
    that are safe to compare or hash.

    Arguments:
        x: the float array to be converted
        where: specifies whether the fractional part (True) or the full value (False)
            of the input is to be used. Must be broadcastable to `x.shape`.
            If False, the output is the same as that of `comparable`.
        bin_density: the inverse width of each bin. When binning rational numbers,
            it's best to use a multiple of all expected denominators `bin_density`.
            The default is :math:`3326400 = 2^6\times 3^3\times 5^2\times 7\times 11`.
        offset: constant offset added to `bin_density * x` before rounding. To minimise
            the chances of "interesting" numbers appearing on bin boundaries, it's
            best to use a rational number with a large prime denominator.
            The default is 5413/15629, both are primes.

    Returns:
        [`x` or frac(`x`)]` * bin_density + offset` rounded to an integer

    Example:

        >>> comparable_periodic([0.0, 0.3, 0.30000001, 1.3], where = [[True], [False]])
        array([[      0,  997920,  997920,  997920],
               [      0,  997920,  997920, 4324320]])
    """
    bins = np.asarray(np.rint(np.asarray(x) * bin_density + offset), dtype=int)
    return np.where(where, bins % bin_density, bins)


def _prune_zeros(x: Array, atol: float = 1e-08) -> Array:
    # prunes nearly zero entries
    x[np.isclose(x, 0.0, rtol=0.0, atol=atol)] = 0.0
    return x


def prune_zeros(x: Array, atol: float = 1e-08) -> Array:
    """Prunes nearly zero real and imaginary parts"""
    if np.iscomplexobj(x):
        # Check if complex part is nonzero at all
        if np.allclose(x.imag, 0.0, rtol=0.0, atol=atol):
            return _prune_zeros(x.real)
        else:
            return _prune_zeros(x.real) + 1j * _prune_zeros(x.imag)
    else:
        return _prune_zeros(x)


def is_approx_int(x: Array, atol: float = 1e-08) -> Array:
    """
    Returns `True` for all elements of the array that are within
    `atol` to an integer.
    """
    return np.isclose(x, np.rint(x), rtol=0.0, atol=atol)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import constraint
from . import index

from .abstract_hilbert import AbstractHilbert
from .discrete_hilbert import DiscreteHilbert
from .homogeneous import HomogeneousHilbert

from .continuous_hilbert import ContinuousHilbert

from .doubled_hilbert import DoubledHilbert
from .spin import Spin
from .fock import Fock
from .qubit import Qubit
from .particle import Particle

from .tensor_hilbert import TensorHilbert
from . import tensor_hilbert_discrete

from . import random


# Deprecated bindings
from .custom_hilbert import CustomHilbert as _deprecated_CustomHilbert

_deprecations = {
    # September 2024, NetKet 3.14
    "CustomHilbert": (
        "netket.hilbert.CustomHilbert is deprecated: use custom constraints with "
        "existing hilbert spaces instead, or define your own hilbert space class.",
        _deprecated_CustomHilbert,
    ),
}

from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr
from netket.utils import _hide_submodules

__getattr__ = _deprecation_getattr(__name__, _deprecations)
_hide_submodules(__name__)

del _deprecation_getattr



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp

from netket.utils import random_seed, mpi, config
from netket.utils.mpi import MPI_jax_comm
from netket.utils.types import PRNGKeyT, SeedT


def PRNGKey(seed: SeedT | None = None, *, root: int = 0, comm=MPI_jax_comm) -> PRNGKeyT:
    """
    Initialises a PRNGKey using an optional starting seed.
    The same seed will be distributed to all processes.
    """
    if seed is None:
        seed = random_seed()

    if isinstance(seed, int):
        # We can't sync the PRNGKey, so we can only sinc integer seeds
        # see https://github.com/google/jax/pull/16511
        if config.netket_experimental_sharding and jax.process_count() > 1:  # type: ignore[attr-defined]
            # TODO: use stable jax function
            from jax.experimental import multihost_utils

            seed = int(
                multihost_utils.broadcast_one_to_all(
                    seed, is_source=jax.process_index() == root
                ).item()
            )

        key = jax.random.PRNGKey(seed)
    else:
        key = seed

    if not config.netket_experimental_sharding:  # type: ignore[attr-defined]
        key = _bcast_key(key, root=root, comm=comm)
    return key


class PRNGSeq:
    """
    A sequence of PRNG keys generated based on an initial key.
    """

    def __init__(self, base_key: SeedT | None = None):
        if base_key is None:
            base_key = PRNGKey()
        elif isinstance(base_key, int):
            base_key = PRNGKey(base_key)
        self._current: PRNGKeyT = base_key

    def __iter__(self):
        return self

    def __next__(self):
        self._current = jax.random.split(self._current, num=1)[0]
        return self._current

    def next(self):
        return self.__next__()

    def take(self, num: int):
        """
        Returns an array of `num` PRNG keys and advances the iterator accordingly.
        """
        keys = jax.random.split(self._current, num=num + 1)
        self._current = keys[-1]
        return keys[:-1]


def mpi_split(key, *, root=0, comm=MPI_jax_comm) -> PRNGKeyT:
    """
    Split a key across MPI nodes in the communicator.
    Only the input key on the root process matters.

    Arguments:
        key: The key to split. Only considered the one on the root process.
        root: (default=0) The root rank from which to take the input key.
        comm: (default=MPI.COMM_WORLD) The MPI communicator.

    Returns:
        A PRNGKey depending on rank number and key.
    """

    # Maybe add error/warning if in_key is not the same
    # on all MPI nodes?
    keys = jax.random.split(key, mpi.n_nodes)

    keys = _bcast_key(keys, root=root, comm=comm)

    return keys[mpi.rank]


def _bcast_key(key, root=0, comm=MPI_jax_comm) -> PRNGKeyT:
    """
    Utility function equivalent to calling `mpi_bcast_jax` on a jax key,
    but working around some sharding bug when not using sharding, arising
    from MPI.
    """
    is_new_style_key = jnp.issubdtype(key.dtype, jax.dtypes.prng_key)

    if is_new_style_key:
        _impl = jax.random.key_impl(key)
        key = jax.random.key_data(key)

    key = jax.tree_util.tree_map(
        lambda k: mpi.mpi_bcast_jax(k, root=root, comm=comm)[0], key
    )

    if is_new_style_key:
        key = jax.random.wrap_key_data(key, impl=_impl)  # type: ignore[arg-type]

    return key


def batch_choice(key, a, p):
    """
    Batched version of `jax.random.choice`.

    Attributes:
      key: a PRNGKey used as the random key.
      a: 1D array. Random samples are generated from its elements.
      p: 2D array of shape `(batch_size, a.size)`. Each slice `p[i, :]` is
        the probabilities associated with entries in `a` to generate a sample
        at the index `i` of the output. Can be unnormalized.

    Returns:
      The generated samples as an 1D array of shape `(batch_size,)`.
    """
    p_cumsum = p.cumsum(axis=1)
    r = p_cumsum[:, -1:] * jax.random.uniform(key, shape=(p.shape[0], 1))
    indices = (r > p_cumsum).sum(axis=1)
    out = a[indices]
    return out



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import netket as nk
from netket.operator import AbstractOperator

from netket.operator._abstract_observable import AbstractObservable


class VarianceObservable(AbstractObservable):
    r"""
    Observable computing the variance of a quantum operator :math:`O`.
    """

    def __init__(self, operator: AbstractOperator, use_Oloc_squared: bool = False):
        r"""
        Constructs the observable computing the variance of an arbitrary quantum operator :math:`O` as:

        .. math::

            \text{Var} = \frac{\langle \Psi | O^2 | \Psi \rangle}{\langle \Psi | \Psi \rangle} - \bigg( \frac{\langle \Psi | O | \Psi \rangle}{\langle \Psi | \Psi \rangle}\bigg)^2

        It can compute the first term using either the estimator of the squared operator :math:`O^2` (more precise but less efficient, since it requires
        the connected configurations and the matrix elements of :math:`O^2`):

        .. math::

            \text{Var} = \mathbb{E}_{\sigma \sim |\Psi(\sigma)|^2}\bigg[\frac{\langle \sigma | O^2 | \Psi \rangle}{\langle \sigma | \Psi \rangle}\bigg] - \bigg(\mathbb{E}_{\sigma \sim |\Psi(\sigma)|^2}\bigg[\frac{\langle \sigma | O | \Psi \rangle}{\langle \sigma | \Psi \rangle}\bigg]\bigg)^2.

        or using the square modulus of the estimator of :math:`O` (more noisy but more efficient):

        .. math::

            \text{Var} = \mathbb{E}_{\sigma \sim |\Psi(\sigma)|^2}\bigg[\bigg(\frac{\langle \sigma | O | \Psi \rangle}{\langle \sigma | \Psi \rangle} - \mathbb{E}_{\sigma \sim |\Psi(\sigma)|^2}\bigg[\frac{\langle \sigma | O | \Psi \rangle}{\langle \sigma | \Psi \rangle}\bigg]\bigg)^2\bigg]

        This VariationalOperator wraps an operator such that the gradient will not be computed
        with respect to the expectation value, but with respect to the variance.

        Args:
            operator: The operator for which the variance is to be computed.
            use_Oloc_squared: (Defaults False) if True, uses the fast estimator obtained by squaring the local estimator O_loc. If False, uses the straightforward estimator of (O@O)_loc which is quadratically more expensive. The fast estimator can sometimes lead to worse quality results.

        Returns:
            Observable computing the variance of `operator`.
        """
        super().__init__(operator.hilbert)
        self._operator = operator

        if use_Oloc_squared:
            self._operator_squared = nk.operator.Squared(operator)
        else:
            self._operator_squared = operator @ operator

    @property
    def operator(self) -> AbstractOperator:
        """
        The operator for which the variance is to be computed.
        """
        return self._operator

    @property
    def operator_squared(self) -> AbstractOperator:
        """
        The squared of the operator for which the variance is to be computed.
        Depending on the flag `use_Oloc_squared`, this can be the operator using the local
        estimator of `O^2` (False), or the one using the square modulus of the
        local estimator of `O` (True).
        """
        return self._operator_squared

    def __repr__(self):
        return f"VarianceObservable(op={self.operator})"



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains functions generating the numba-packed representation of local
operators.
"""

import numpy as np
import numba
from scipy import sparse
import jax.numpy as jnp

from netket.hilbert import AbstractHilbert
from netket.utils.types import DType


def pack_internals(
    hilbert: AbstractHilbert,
    operators_dict: dict,
    constant,
    dtype: DType,
    mel_cutoff: float,
):
    """
    Take the internal lazy representation of a local operator and returns the arrays
    needed for the numba implementation.

    This takes as input a dictionary with Tuples as keys, the `acting_on` and matrices as values.
    The keys represent the sites upon which the matrix acts.
    It is assumed that the integer in the tuples are sorted.

    Returns a dictionary with all the data fields
    """
    op_acting_on = list(operators_dict.keys())
    operators = list(operators_dict.values())
    n_operators = len(operators_dict)

    """Analyze the operator strings and precompute arrays for get_conn inference"""

    # how many sites each operator is acting on
    acting_size = np.array([len(aon) for aon in op_acting_on], dtype=np.intp)

    # compute the maximum number of off-diagonal nonzeros (over all rows) of each operator
    op_n_conns_offdiag = max_nonzero_per_row(operators, mel_cutoff)

    # Support empty LocalOperators such as the identity.
    if len(acting_size) > 0:
        # maximum number of sites any operator is acting on
        max_acting_on_sz = np.max(acting_size)

        # maximum size of any operator (maximum size of the matrix / prod of local hilbert spaces)
        max_op_size = max(map(lambda x: x.shape[0], operators))
        # maximum number of off-diagonal nonzeros of any operator
        max_op_size_offdiag = np.max(op_n_conns_offdiag)
    else:
        max_acting_on_sz = 0
        max_op_size = 0
        max_op_size_offdiag = 0

    # matrix storing which sites each operator acts on, padded with -1
    acting_on = np.full((n_operators, max_acting_on_sz), -1, dtype=np.intp)
    for i, aon in enumerate(op_acting_on):
        acting_on[i][: len(aon)] = aon

    ###
    # allocate empty arrays which are filled below

    # array storing the basis for each site each operator is acting on
    # The basis is an integer used to map (indices of) local states on sites
    # to (indices of) states in the space spanned by all sites the op is acting on
    # (it is simply the product of number of local states of the sites before)
    basis = np.full((n_operators, max_acting_on_sz), 0x7FFFFFFF, dtype=np.int64)

    diag_mels = np.full((n_operators, max_op_size), np.nan, dtype=dtype)

    mels = np.full(
        (n_operators, max_op_size, max_op_size_offdiag),
        np.nan,
        dtype=dtype,
    )
    # x_prime contains the local state after the operator has been applied
    # for the sites the operator is acting on, for each row
    # (each row also corresponds to a list of local states)
    x_prime = np.full(
        (n_operators, max_op_size, max_op_size_offdiag, max_acting_on_sz),
        -1,
        dtype=np.float64,
    )
    # store the number of off-diagonal nonzeros per row of each operator
    n_conns = np.full((n_operators, max_op_size), 0, dtype=np.intp)

    ###
    # iterate over all operators
    for i, (aon, op) in enumerate(operators_dict.items()):
        # how many sites this operator is acting on
        aon_size = len(aon)

        n_local_states_per_site = np.asarray([hilbert.size_at_index(i) for i in aon])

        # compute the basis of each site of this operator
        # i.e. the product of the number of local states of all sites before it
        ba = 1
        for s in range(aon_size):
            basis[i, s] = ba
            ba *= hilbert.shape[aon[aon_size - s - 1]]

        if sparse.issparse(op):
            if not isinstance(op, sparse.csr_matrix):
                op = op.tocsr()
            # Extract the sparse matrix representation to numpy arrays
            data = np.array(op.data, copy=False)
            indices = np.array(op.indices, copy=False)
            indptr = np.array(op.indptr, copy=False)

            _append_matrix_sparse(
                data,
                indices,
                indptr,
                aon_size,
                n_local_states_per_site,
                mel_cutoff,
                diag_mels[i],
                mels[i],
                x_prime[i],
                n_conns[i],
            )

        else:
            _append_matrix(
                op,
                aon_size,
                n_local_states_per_site,
                mel_cutoff,
                diag_mels[i],
                mels[i],
                x_prime[i],
                n_conns[i],
            )

    nonzero_diagonal = (
        np.any(np.abs(diag_mels) >= mel_cutoff) or np.abs(constant) >= mel_cutoff
    )

    max_conn_size = 1 if nonzero_diagonal else 0
    # estimate max_conn_size with the sum of the
    # maximum number of off-diagonal nonzeros of all operators
    max_conn_size = max_conn_size + np.sum(op_n_conns_offdiag)

    return {
        "acting_on": acting_on,
        "acting_size": acting_size,
        "diag_mels": diag_mels,
        "mels": mels,
        "x_prime": x_prime,
        "n_conns": n_conns,
        "basis": basis,
        "nonzero_diagonal": nonzero_diagonal,
        "max_conn_size": max_conn_size,
    }


def pack_internals_jax(
    hilbert: AbstractHilbert,
    operators_dict: dict,
    constant,
    dtype: DType,
    mel_cutoff: float,
):
    # we groups together operators which act on the same number of sites
    # and then call pack_internals on each group
    #
    # TODO in the future consider separating also operators with different sizes
    # to avoid excessive padding
    # (only relevant for non-uniform number of local states)

    op_acting_on = list(operators_dict.keys())
    operators = list(operators_dict.values())
    # how many sites each operator is acting on
    acting_size = np.array([len(aon) for aon in op_acting_on], dtype=np.intp)

    data = {}
    data["nonzero_diagonal"] = np.abs(constant) >= mel_cutoff
    data["max_conn_size"] = 0

    # iterate over groups of operators with same number of sites
    # special case for empty operator
    for s in np.unique(acting_size) if len(acting_size) > 0 else [0]:
        (indices,) = np.where(acting_size == s)
        operators_dict_s = {op_acting_on[i]: operators[i] for i in indices}
        data_s = pack_internals(hilbert, operators_dict_s, 0, dtype, mel_cutoff)
        nonzero_diagonal = bool(data_s.pop("nonzero_diagonal"))
        max_conn_size = int(data_s.pop("max_conn_size"))
        if nonzero_diagonal:
            # we only count the diag elem once below, so we subtract it here
            max_conn_size = max_conn_size - 1
        data["nonzero_diagonal"] = data["nonzero_diagonal"] or nonzero_diagonal
        data["max_conn_size"] = data["max_conn_size"] + max_conn_size
        # append other elements to lists:
        for k, v in data_s.items():
            data[k] = data.pop(k, []) + [jnp.asarray(v)]

    # count the diagonal once at the end
    if data["nonzero_diagonal"]:
        data["max_conn_size"] = data["max_conn_size"] + 1

    return data


@numba.jit(nopython=True)
def _append_matrix(
    operator,
    acting_size,
    hilb_size_per_site,
    epsilon,
    diag_mels,
    mels,
    x_prime,
    n_conns,
):
    """
    Appends
    """
    op_size = operator.shape[0]
    assert op_size == operator.shape[1]
    # iterate over rows
    for i in range(op_size):
        # set diag mels
        diag_mels[i] = operator[i, i]
        # count number of connected elements
        n_conns[i] = 0  # k_conn = 0
        # iterate over cols
        for j in range(op_size):
            # off-diagonal, non-zero
            if i != j and np.abs(operator[i, j]) > epsilon:
                k_conn = n_conns[i]
                # set off-doagnoal mels
                mels[i, k_conn] = operator[i, j]
                # convert the col to the corresponding local states
                # and store it in x_prime
                _number_to_state(
                    j,
                    hilb_size_per_site,
                    x_prime[i, k_conn, :acting_size],
                )
                n_conns[i] += 1  # k_conn=k_conn+1


@numba.jit(nopython=True)
def _append_matrix_sparse(
    data,
    indices,
    indptr,
    acting_size,
    hilb_size_per_site,
    epsilon,
    diag_mels,
    mels,
    x_prime,
    n_conns,
):
    """
    Equivalent to _append_matrix, but takes as input the three arrays
    'data, indices, indptr' of the CSR sparse format instead of a numpy
    dense matrix.
    """
    op_size = len(indptr) - 1

    for i in range(op_size):
        # If the diagonal element was not found in the data, set it to 0
        diag_mels[i] = 0
        for index in range(indptr[i], indptr[i + 1]):
            j = indices[index]
            val = data[index]

            if i == j:  # Diagonal elements
                diag_mels[i] = val

            elif np.abs(val) > epsilon:  # Non-diagonal elements
                k_conn = n_conns[i]
                mels[i, k_conn] = val
                _number_to_state(
                    j,
                    hilb_size_per_site,
                    x_prime[i, k_conn, :acting_size],
                )
                n_conns[i] += 1


@numba.jit(nopython=True)
def _number_to_state(number, hilbert_size_per_site, out):
    out[:] = 0
    size = out.shape[0]

    ip = number
    k = size - 1
    while ip > 0:
        local_size = hilbert_size_per_site[k]
        out[k] = ip % local_size
        ip = ip // local_size
        k -= 1

    return out


# Generated by chatting with chatgpt
# https://chat.openai.com/share/9fb4b0d3-0306-41ca-aefc-eee780a2dc02
def max_nonzero_per_row(operators, cutoff):
    """
    This function counts the maximum number of nonzero entries per row, excluding the
    diagonal, among all operators in a local operator.

    Prior to this function, netket defaulted to returning the maximum shape of the
    operators minus 1, but that was largely memory inefficient.

    This function has a computational overhead, as we have to iterate through all
    operators, but it allows us to be much happier in terms of memory cost.
    """
    max_counts = []
    for matrix in operators:
        # Check if the matrix is sparse
        if sparse.issparse(matrix):
            # simple implementation, raises warning
            # matrix = matrix.copy()  # Ensure we don't modify the original matrix
            # matrix.data[np.abs(matrix.data) < cutoff] = 0  # Apply cutoff
            # matrix.setdiag(0)  # Set diagonal entries to zero

            # alternative implementation. Does not raise warning but
            # I'm unsure if it's any faster...
            # Convert to dok_matrix format for efficient modification
            dok_matrix = matrix.todok()
            dok_matrix.setdiag(0)

            # Collect keys to delete
            keys_to_delete = [
                (i, j) for (i, j), v in dok_matrix.items() if abs(v) < cutoff
            ]

            # Delete keys
            for key in keys_to_delete:
                del dok_matrix[key]

            row_counts = dok_matrix.tocsr().getnnz(
                axis=1
            )  # Count non-zero entries in each row
        else:
            matrix = matrix.copy()  # Make a copy to avoid modifying the original matrix
            mask = np.abs(matrix) >= cutoff  # Create a mask of entries above the cutoff
            np.fill_diagonal(mask, 0)  # Set diagonal entries to zero
            row_counts = np.count_nonzero(
                mask, axis=1
            )  # Count non-zero entries in each row

        max_counts.append(np.max(row_counts))
    return np.array(max_counts, dtype=np.int32)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from functools import wraps

import jax
from jax import numpy as jnp

from jax.nn import celu
from jax.nn import elu
from jax.nn import gelu
from jax.nn import glu
from jax.nn import leaky_relu

from jax.nn import log_sigmoid
from jax.nn import log_softmax
from jax.nn import standardize
from jax.nn import relu

from jax.nn import sigmoid
from jax.nn import soft_sign
from jax.nn import softmax

from jax.nn import softplus
from jax.nn import swish
from jax.nn import silu
from jax.nn import selu
from jax.nn import hard_tanh
from jax.nn import relu6
from jax.nn import hard_sigmoid
from jax.nn import hard_swish

from jax.numpy import tanh
from jax.numpy import cosh
from jax.numpy import sinh


from netket.jax import HashablePartial
from netket.utils import deprecated_new_name as _deprecated_new_name


def reim(f):
    r"""Modifies a non-linearity to act separately on the real and imaginary parts"""

    @wraps(f)
    def reim_f(f, x):
        sqrt2 = jnp.sqrt(jnp.array(2, dtype=x.real.dtype))
        if jnp.iscomplexobj(x):
            return jax.lax.complex(f(sqrt2 * x.real), f(sqrt2 * x.imag)) / sqrt2
        else:
            return f(x)

    fun = HashablePartial(reim_f, f)

    fun.__name__ = f"reim_{f.__name__}"  # type: ignore
    fun.__doc__ = (
        f"{f.__name__} applied separately to the real and"
        f"imaginary parts of it's input.\n\n"
        f"The docstring to the original function follows.\n\n"
        f"{f.__doc__}"
    )

    return fun


def log_cosh(x):
    """
    Logarithm of the hyperbolic cosine, implemented in a more stable way.
    """
    sgn_x = -2 * jnp.signbit(x.real) + 1
    x = x * sgn_x
    return x + jnp.log1p(jnp.exp(-2.0 * x)) - jnp.log(2.0)


def log_sinh(x):
    """
    Logarithm of the hyperbolic sine.
    """
    return jax.numpy.log(jax.numpy.sinh(x))


def log_tanh(x):
    """
    Logarithm of the hyperbolic tangent.
    """
    return jax.numpy.log(jax.numpy.tanh(x))


reim_selu = reim(selu)
r"""Returns the selu non-linearity, applied separately to the real and imaginary parts"""

reim_relu = reim(relu)
r"""Returns the relu non-linearity, applied separately to the real and imaginary parts"""


# TODO: Deprecated in January 2024, remove in 2025
@_deprecated_new_name("standardize", reason="Because it is deprecated by jax as well")
def normalize(*args, **kwargs):
    return standardize(*args, **kwargs)



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
from collections.abc import Iterable


import rich


def __repr_from_rich__(self):
    """
    default __repr__ that calls __rich__
    """
    console = rich.get_console()
    with console.capture() as capture:
        console.print(self, end="")
    return capture.get()


def _repr_mimebundle_from_rich_(
    self, include: Iterable[str], exclude: Iterable[str], **kwargs: Any
) -> dict[str, str]:
    from rich.jupyter import _render_segments

    console = rich.get_console()
    segments = list(console.render(self, console.options))  # type: ignore
    html = _render_segments(segments)
    text = console._render_buffer(segments)
    data = {"text/plain": text, "text/html": html}
    if include:
        data = {k: v for (k, v) in data.items() if k in include}
    if exclude:
        data = {k: v for (k, v) in data.items() if k not in exclude}
    return data


def rich_repr(clz):
    """
    Class decorator setting the repr method to use
    `rich`.
    """
    setattr(clz, "__repr__", __repr_from_rich__)
    setattr(clz, "_repr_mimebundle_", _repr_mimebundle_from_rich_)
    return clz



# Copyright 2021 The NetKet Authors - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional
from functools import partial

import jax
import jax.numpy as jnp

from netket.utils.numbers import dtype as _dtype
from netket.utils.mpi.primitives import mpi_all_jax
from netket.utils import struct
from netket.utils.types import PyTree

from ._integrator_state import IntegratorState, IntegratorFlags
from ._integrator_params import IntegratorParameters
from ._solver import AbstractSolver
from ._utils import (
    maybe_jax_jit,
    scaled_error,
    propose_time_step,
    set_flag_jax,
    euclidean_norm,
    maximum_norm,
)


class Integrator(struct.Pytree, mutable=True):
    r"""
    Ordinary-Differential-Equation integrator.
    Given an ODE-function :math:`dy/dt = f(t, y)`, it integrates the derivatives to obtain the solution
    at the next time step :math:`y_{t+1}`.
    """

    f: Callable = struct.field(pytree_node=False)
    """The ODE function."""

    _state: IntegratorState
    """The state of the integrator, containing informations about the solution."""
    _solver: AbstractSolver
    """The ODE solver."""
    _parameters: IntegratorParameters
    """The options of the integration."""

    use_adaptive: bool = struct.field(pytree_node=False)
    """Boolean indicating whether to use an adaptative scheme."""
    norm: Callable = struct.field(pytree_node=False)
    """The norm used to estimate the error."""

    def __init__(
        self,
        f: Callable,
        solver: AbstractSolver,
        t0: float,
        y0: PyTree,
        use_adaptive: bool,
        parameters: IntegratorParameters,
        norm: str | Callable = None,
    ):
        r"""
        Args:
            f: The ODE function
                Given a time `t` and a state `y_t`, it should return the partial
                derivatives in the same format as `y_t`. The function should also accept
                supplementary arguments, such as :code:`stage`.
            solver: The ODE solver.
                If :code:`use_adaptive`, it should define the method :code:`step_with_error(f,dt,t,y_t,solver_state)`
                otherwise, the method :code:`step(f,dt,t,y_t,solver_state)`
            t0: The intial time.
            y0: The initial state.
            use_adaptive: The (boolean) indicator whether an adaptive (time-step) scheme is used.
            parameters: The suppleementary hyper-parameters of the integrator.
                This includes the values for :code:`dt`, :code:`atol`, :code`rtol` and :code`dt_limits`
                See :code:`IntegratorParameters` for more details.
            norm: The function used for the norm of the error.
                By default, we use euclidean_norm.
        """
        self.f = f
        self._solver = solver
        self._parameters = parameters

        self.use_adaptive = use_adaptive

        if norm is not None:
            if isinstance(norm, str):
                norm = norm.lower()
                if norm == "euclidean":
                    norm = euclidean_norm
                elif norm == "maximum":
                    norm = maximum_norm
                else:
                    raise ValueError(
                        f"The error norm must either be 'euclidean' or 'maximum', instead got {norm}."
                    )
            if not isinstance(norm, Callable):
                raise ValueError(
                    f"The error norm must be a callable, instead got a {type(norm)}."
                )
        else:
            norm = euclidean_norm
        self.norm = norm

        self._state = self._init_state(t0, y0)

    def _init_state(self, t0: float, y0: PyTree) -> IntegratorState:
        r"""
        Initializes the `IntegratorState` structure containing the solver and state,
        given the necessary information.

        Args:
            t0: The initial time of evolution
            y0: The solution at initial time `t0`

        Returns:
            An :code:`Integrator` instance intialized with the passed arguments
        """
        dt = self._parameters.dt

        t_dtype = jnp.result_type(_dtype(t0), _dtype(dt))

        return IntegratorState(
            t=jnp.array(t0, dtype=t_dtype),
            y=y0,
            dt=jnp.array(dt, dtype=t_dtype),
            solver=self._solver,
            last_norm=0.0 if self.use_adaptive else None,
            last_scaled_error=0.0 if self.use_adaptive else None,
            flags=IntegratorFlags(0),
        )

    def step(self, max_dt: float = None) -> bool:
        """
        Performs one full step by :code:`min(self.dt, max_dt)`.

        Returns:
            A boolean indicating whether the step was successful or
            was rejected by the step controller and should be retried.

            Note that the step size can be adjusted by the step controller
            in both cases, so the integrator state will have changed
            even after a rejected step.
        """
        if not self.use_adaptive:
            self._state = self._step_fixed(
                solver=self._solver,
                f=self.f,
                state=self._state,
                max_dt=max_dt,
                parameters=self._parameters,
            )
        else:
            self._state = self._step_adaptive(
                solver=self._solver,
                f=self.f,
                state=self._state,
                max_dt=max_dt,
                parameters=self._parameters,
                norm_fn=self.norm,
            )

        return self._state.accepted

    @property
    def t(self) -> float:
        """The actual time."""
        return self._state.t.value

    @property
    def y(self) -> PyTree:
        """The actual state."""
        return self._state.y

    @property
    def dt(self) -> float:
        """The actual time-step size."""
        return self._state.dt

    @property
    def solver(self) -> AbstractSolver:
        """The ODE solver."""
        return self._solver

    def _get_integrator_flags(self, intersect=IntegratorFlags.NONE) -> IntegratorFlags:
        r"""Returns the currently set flags of the integrator, intersected with `intersect`."""
        # _state.flags is turned into an int-valued DeviceArray by JAX,
        # so we convert it back.
        return IntegratorFlags(int(self._state.flags) & intersect)

    @property
    def errors(self) -> IntegratorFlags:
        r"""Returns the currently set error flags of the integrator."""
        return self._get_integrator_flags(IntegratorFlags.ERROR_FLAGS)

    @property
    def warnings(self) -> IntegratorFlags:
        r"""Returns the currently set warning flags of the integrator."""
        return self._get_integrator_flags(IntegratorFlags.WARNINGS_FLAGS)

    def __repr__(self) -> str:
        return "{}(solver={}, state={}, adaptive={}{})".format(
            "Integrator",
            self.solver,
            self._state,
            self.use_adaptive,
            (f", norm={self.norm}" if self.use_adaptive else ""),
        )

    @staticmethod
    @partial(maybe_jax_jit, static_argnames=["f"])
    def _step_fixed(
        solver: AbstractSolver,
        f: Callable,
        state: IntegratorState,
        max_dt: Optional[float],
        parameters: IntegratorParameters,
    ) -> IntegratorState:
        r"""
        Performs one fixed step from current time.
        Args:
            solver: Instance that solves the ODE.
                The solver should contain a method :code:`step(f,dt,t,y_t,solver_state)`
            f: A callable ODE function.
                Given a time `t` and a state `y_t`, it should return the partial
                derivatives in the same format as `y_t`. The function should also accept
                supplementary arguments, such as :code:`stage`.
            state: IntegratorState containing the current state (t,y), the solver_state and stability information.
            max_dt: The maximal value for the time step `dt`.
            parameters: The integration parameters.

        Returns:
            Updated state of the integrator.
        """
        del parameters

        if max_dt is None:
            actual_dt = state.dt
        else:
            actual_dt = jnp.minimum(state.dt, max_dt)

        # Perform the solving step
        y_tp1, solver_state = solver.step(
            f, actual_dt, state.t.value, state.y, state.solver_state
        )

        return state.replace(
            step_no=state.step_no + 1,
            step_no_total=state.step_no_total + 1,
            t=state.t + actual_dt,
            y=y_tp1,
            solver_state=solver_state,
            flags=IntegratorFlags.INFO_STEP_ACCEPTED,
        )

    @staticmethod
    @partial(maybe_jax_jit, static_argnames=["f", "norm_fn"])
    def _step_adaptive(
        solver: AbstractSolver,
        f: Callable,
        state: IntegratorState,
        max_dt: Optional[float],
        parameters: IntegratorParameters,
        norm_fn: Callable,
    ) -> IntegratorState:
        r"""
        Performs one adaptive step from current time.
        Args:
            solver: Instance that solves the ODE
                The solver should contain a method :code:`step_with_error(f,dt,t,y_t,solver_state)`
            f: A callable ODE function.
                Given a time `t` and a state `y_t`, it should return the partial
                derivatives in the same format as `y_t`. The function should also accept
                supplementary arguments, such as :code:`stage`.
            state: IntegratorState containing the current state (t,y), the solver_state and stability information.
            norm_fn: The function used for the norm of the error.
                By default, we use euclidean_norm.
            parameters: The integration parameters.
            max_dt: The maximal value for the time-step size `dt`.

        Returns:
            Updated state of the integrator
        """
        flags = IntegratorFlags(0)

        if max_dt is None:
            actual_dt = state.dt
        else:
            actual_dt = jnp.minimum(state.dt, max_dt)

        # Perform the solving step
        y_tp1, y_err, solver_state = solver.step_with_error(
            f, actual_dt, state.t.value, state.y, state.solver_state
        )

        scaled_err, norm_y = scaled_error(
            y_tp1,
            y_err,
            parameters.atol,
            parameters.rtol,
            last_norm_y=state.last_norm,
            norm_fn=norm_fn,
        )

        # Propose the next time step, but limited within [0.1 dt, 5 dt] and potential
        # global limits in dt_limits. Not used when actual_dt < state.dt (i.e., the
        # integrator is doing a smaller step to hit a specific stop).
        dt_min, dt_max = parameters.dt_limits
        next_dt = propose_time_step(
            actual_dt,
            scaled_err,
            solver.error_order,
            limits=(
                (jnp.maximum(0.1 * state.dt, dt_min) if dt_min else 0.1 * state.dt),
                (jnp.minimum(5.0 * state.dt, dt_max) if dt_max else 5.0 * state.dt),
            ),
        )

        # check if next dt is NaN
        flags = set_flag_jax(
            ~jnp.isfinite(next_dt), flags, IntegratorFlags.ERROR_INVALID_DT
        )

        # check if we are at lower bound for dt
        if dt_min is not None:
            is_at_min_dt = jnp.isclose(next_dt, dt_min)
            flags = set_flag_jax(is_at_min_dt, flags, IntegratorFlags.WARN_MIN_DT)
        else:
            is_at_min_dt = False
        if dt_max is not None:
            is_at_max_dt = jnp.isclose(next_dt, dt_max)
            flags = set_flag_jax(is_at_max_dt, flags, IntegratorFlags.WARN_MAX_DT)

        # accept if error is within tolerances or we are already at the minimal step
        accept_step = jnp.logical_or(scaled_err < 1.0, is_at_min_dt)
        # accept the time step iff it is accepted by all MPI processes
        accept_step, _ = mpi_all_jax(accept_step)

        return jax.lax.cond(
            accept_step,
            # step accepted
            lambda _: state.replace(
                step_no=state.step_no + 1,
                step_no_total=state.step_no_total + 1,
                y=y_tp1,
                t=state.t + actual_dt,
                dt=jax.lax.cond(
                    actual_dt == state.dt,
                    lambda _: next_dt,
                    lambda _: state.dt,
                    None,
                ),
                last_norm=norm_y.astype(state.last_norm.dtype),
                last_scaled_error=scaled_err.astype(state.last_scaled_error.dtype),
                solver_state=solver_state,
                flags=flags | IntegratorFlags.INFO_STEP_ACCEPTED,
            ),
            # step rejected, repeat with lower dt
            lambda _: state.replace(
                step_no_total=state.step_no_total + 1,
                dt=next_dt,
                flags=flags,
            ),
            state,
        )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable
from functools import partial


from flax.core import FrozenDict

from ..dispatch import dispatch

from .history import History

AccumulatorFun = Callable[[Any, Any], Any]

AccumulatorFunTypeRegistry: dict[type, AccumulatorFun] = {}
"""
Dictionary used for dispatching the accumulator functino based on the type
of the input tree.
"""


def _accum_histories_container(
    fun: Callable[[Any, Any], Any],
    tree_accum: list | Any,
    tree: list | Any,
    container_type: type = list,
    **kwargs,
):
    # If nothing to accumulate, just return the former tree
    if len(tree) == 0:
        return tree_accum

    if tree_accum is None:
        tree_accum = [None for _ in range(len(tree))]
    return container_type(
        accum_in_tree(fun, _accum, _tree, **kwargs)
        for _accum, _tree in zip(tree_accum, tree)
    )


def _accum_histories_dict(
    fun: Callable[[Any, Any], Any],
    tree_accum: list | Any,
    tree: list,
    **kwargs,
):
    # If nothing to accumulate, just return the former tree
    if len(tree.keys()) == 0:
        return tree_accum

    if tree_accum is None:
        tree_accum = {}
    for key in tree.keys():
        value = accum_in_tree(fun, tree_accum.get(key, None), tree[key], **kwargs)
        #
        if value is not None:
            tree_accum[key] = value
    return tree_accum


AccumulatorFunTypeRegistry[list] = partial(
    _accum_histories_container, container_type=list
)
AccumulatorFunTypeRegistry[tuple] = partial(
    _accum_histories_container, container_type=tuple
)
AccumulatorFunTypeRegistry[dict] = _accum_histories_dict
AccumulatorFunTypeRegistry[FrozenDict] = _accum_histories_dict


def accum_in_tree(
    fun: Callable[[Any, Any], Any],
    tree_accum: dict[str, Any] | Any,
    tree: dict[str, Any],
    **kwargs,
) -> dict[str, Any]:
    """
    Maps all the leafs in the two trees, applying the function with the leafs of tree1
    as first argument and the leafs of tree2 as second argument
    Any additional argument after the first two is forwarded to the function call.

    This is useful e.g. to sum the leafs of two trees

    Args:
        fun: the function to apply to all leafs
        tree1: the structure containing leafs. This can also be just a leaf
        tree2: the structure containing leafs. This can also be just a leaf
        *args: additional positional arguments passed to fun
        **kwargs: additional kw arguments passed to fun

    Returns:
        An equivalent tree, containing the result of the function call.
    """
    if tree is None:
        return tree_accum

    tree_type = type(tree)
    if tree_type in AccumulatorFunTypeRegistry:
        return AccumulatorFunTypeRegistry[tree_type](fun, tree_accum, tree, **kwargs)
    elif hasattr(tree, "to_compound"):
        return fun(tree_accum, tree, **kwargs)
    elif hasattr(tree, "to_dict"):
        return accum_in_tree(fun, tree_accum, tree.to_dict(), **kwargs)
    else:
        return fun(tree_accum, tree, **kwargs)


@dispatch
def init_history_from_data(val: Any, step: Any):
    return History(val, step)


def accum_histories(accum: History | None, data, *, step=0) -> History:
    if accum is None:
        return init_history_from_data(data, step)
    else:
        accum.append(data, it=step)
        return accum


accum_histories_in_tree = partial(accum_in_tree, accum_histories)



# Copyright 2020, 2021  The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from functools import partial

import jax

import netket as nk
from netket.jax import tree_cast
from netket.operator import AbstractOperator
from netket.optimizer import LinearOperator
from netket.optimizer.qgt import QGTAuto
from netket.vqs import (
    VariationalState,
    VariationalMixedState,
    MCState,
    FullSumState,
)
from netket.experimental.dynamics import AbstractSolver

from .tdvp_common import TDVPBaseDriver, odefun


class TDVP(TDVPBaseDriver):
    """
    Variational time evolution based on the time-dependent variational principle which,
    when used with Monte Carlo sampling via :class:`netket.vqs.MCState`, is the time-dependent VMC
    (t-VMC) method.

    .. note::
        This TDVP Driver uses the time-integrators from the `nkx.dynamics` module, which are
        automatically executed under a `jax.jit` context.

        When running computations on GPU, this can lead to infinite hangs or extremely long
        compilation times. In those cases, you might try setting the configuration variable
        `nk.config.netket_experimental_disable_ode_jit = True` to mitigate those issues.

    """

    def __init__(
        self,
        operator: AbstractOperator,
        variational_state: VariationalState,
        # TODO: remove default None once `integrator` is removed
        ode_solver: AbstractSolver = None,
        *,
        t0: float = 0.0,
        propagation_type: str = "real",
        # TODO: integrator deprecated in 3.16 (oct/nov 2024)
        integrator: AbstractSolver = None,
        qgt: LinearOperator = None,
        linear_solver=nk.optimizer.solver.pinv_smooth,
        linear_solver_restart: bool = False,
        error_norm: str | Callable = "euclidean",
    ):
        r"""
        Initializes the time evolution driver.

        Args:
            operator: The generator of the dynamics (Hamiltonian for pure states,
                Lindbladian for density operators).
            variational_state: The variational state.
            ode_solver: Solving algorithm used the ODE.
            t0: Initial time at the start of the time evolution.
            propagation_type: Determines the equation of motion: "real"  for the
                real-time Schrödinger equation (SE), "imag" for the imaginary-time SE.
            qgt: The QGT specification.
            linear_solver: The solver for solving the linear system determining the time evolution.
                This must be a jax-jittable function :code:`f(A,b) -> x` that accepts a Matrix-like, Linear Operator
                PyTree object :math:`A` and a vector-like PyTree :math:`b` and returns the PyTree :math:`x` solving
                the system :math:`Ax=b`.
                Defaults to :func:`nk.optimizer.solver.pinv_smooth` with the default svd threshold of 1e-10.
                To change the svd threshold you can use :func:`functools.partial` as follows:
                :code:`functools.partial(nk.optimizer.solver.pinv_smooth, rcond_smooth=1e-5)`.
            linear_solver_restart: If False (default), the last solution of the linear system
                is used as initial value in subsequent steps.
            error_norm: Norm function used to calculate the error with adaptive integrators.
                Can be either "euclidean" for the standard L2 vector norm :math:`w^\dagger w`,
                "maximum" for the maximum norm :math:`\max_i |w_i|`
                or "qgt", in which case the scalar product induced by the QGT :math:`S` is used
                to compute the norm :math:`\Vert w \Vert^2_S = w^\dagger S w` as suggested
                in PRL 125, 100503 (2020).
                Additionally, it possible to pass a custom function with signature
                :code:`norm(x: PyTree) -> float`
                which maps a PyTree of parameters :code:`x` to the corresponding norm.
                Note that norm is used in jax.jit-compiled code.
        """
        if qgt is None:
            qgt = QGTAuto(solver=linear_solver)

        self.propagation_type = propagation_type
        if isinstance(variational_state, VariationalMixedState):
            # assuming Lindblad Dynamics
            # TODO: support density-matrix imaginary time evolution
            if propagation_type == "real":
                self._loss_grad_factor = 1.0
            else:
                raise ValueError(
                    "only real-time Lindblad evolution is supported for " "mixed states"
                )
        else:
            if propagation_type == "real":
                self._loss_grad_factor = -1.0j
            elif propagation_type == "imag":
                self._loss_grad_factor = -1.0
            else:
                raise ValueError("propagation_type must be one of 'real', 'imag'")

        self.qgt = qgt
        self.linear_solver = linear_solver
        self.linear_solver_restart = linear_solver_restart

        super().__init__(
            operator,
            variational_state,
            ode_solver,
            t0=t0,
            error_norm=error_norm,
            integrator=integrator,
        )


@odefun.dispatch
def odefun_tdvp(  # noqa: F811
    state: MCState | FullSumState, driver: TDVP, t, w, *, stage=0
):
    # pylint: disable=protected-access

    state.parameters = w
    state.reset()

    op_t = driver.generator(t)

    driver._loss_stats, driver._loss_forces = state.expect_and_forces(
        op_t,
    )
    driver._loss_grad = _map_parameters(
        driver._loss_forces,
        state.parameters,
        driver._loss_grad_factor,
        driver.propagation_type,
        type(state),
    )

    qgt = driver.qgt(driver.state)
    if stage == 0:  # TODO: This does not work with FSAL.
        driver._last_qgt = qgt

    initial_dw = None if driver.linear_solver_restart else driver._dw
    driver._dw, _ = qgt.solve(driver.linear_solver, driver._loss_grad, x0=initial_dw)

    # If parameters are real, then take only real part of the gradient (if it's complex)
    driver._dw = tree_cast(driver._dw, state.parameters)

    return driver._dw


@partial(jax.jit, static_argnums=(3, 4))
def _map_parameters(forces, parameters, loss_grad_factor, propagation_type, state_T):
    forces = jax.tree_util.tree_map(
        lambda x, target: loss_grad_factor * x,
        forces,
        parameters,
    )

    forces = tree_cast(forces, parameters)

    return forces



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
import warnings

from flax.core.scope import CollectionFilter, DenyList  # noqa: F401

from netket.vqs import expect_and_grad

# TODO:  merged with above once stabilised
from netket.operator._abstract_observable import AbstractObservable

from .state import MCState
from .expect_grad import expect_and_grad_nonhermitian


def ignore_chunk_warning(vstate, operator, chunk_size, name=""):
    return f"""
            Ignoring chunk_size={chunk_size} for {name} method with signature
            ({type(vstate)}, {type(operator)}) because no implementation supporting
            chunking for this signature exists.
            """


# If chunk size is unspecified, set it to None
@expect_and_grad.dispatch
def expect_and_grad_chunking_unspecified(  # noqa: F811
    vstate: MCState,
    operator: AbstractObservable,
    **kwargs,
):
    return expect_and_grad(vstate, operator, None, **kwargs)


# if no implementation exists for batched, run the code unbatched
@expect_and_grad.dispatch(precedence=-10)
def expect_and_grad_fallback(  # noqa: F811
    vstate: MCState,
    operator: AbstractObservable,
    chunk_size: int | tuple,
    *args,
    **kwargs,
):
    warnings.warn(
        ignore_chunk_warning(vstate, operator, chunk_size, name="expect_and_grad")
    )
    return expect_and_grad(vstate, operator, None, *args, **kwargs)


@expect_and_grad_nonhermitian.dispatch(precedence=-10)
def expect_and_grad_nonhermitian_chunk_fallback(
    vstate: MCState,
    Ô,
    chunk_size: Any,
    **kwargs,
):
    warnings.warn(
        ignore_chunk_warning(vstate, Ô, chunk_size, name="expect_and_grad_nonhermitian")
    )
    return expect_and_grad_nonhermitian(vstate, Ô, None, **kwargs)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Ignore false-positives for redefined `product` functions:
# pylint: disable=function-redefined

import itertools

import numpy as np

from netket.utils import HashableArray, struct
from netket.utils.types import Array, DType, Shape
from netket.utils.dispatch import dispatch

from ._group import FiniteGroup
from ._semigroup import Element


class Permutation(Element):
    def __init__(self, permutation: Array, name: str | None = None):
        r"""
        Creates a `Permutation` from an array of preimages of :code:`range(N)`.

        Arguments:
            permutation: 1D array listing :math:`g^{-1}(x)` for all :math:`0\le x < N`
                (i.e., :code:`V[permutation]` permutes the elements of `V` as desired)
            name: optional, custom name for the permutation

        Returns:
            a `Permutation` object encoding the same permutation
        """
        self.permutation = HashableArray(np.asarray(permutation))
        self.__name = name

    def __hash__(self):
        return hash(self.permutation)

    def __eq__(self, other):
        if isinstance(other, Permutation):
            return self.permutation == other.permutation
        else:
            return False

    @property
    def _name(self):
        return self.__name

    def __repr__(self):
        if self._name is not None:
            return self._name
        else:
            return f"Permutation({np.asarray(self).tolist()})"

    def __array__(self, dtype: DType = None):
        return np.asarray(self.permutation, dtype)

    def apply_to_id(self, x: Array):
        """Returns the image of indices `x` under the permutation"""
        return np.argsort(self.permutation)[x]


@dispatch
def product(p: Permutation, x: Array):
    # if p.permutation is a HashableArray and x is a jax Array
    # direct indexing fails, so we call np.asarray on it to extract the
    # wrapped array
    # TODO make indexing work with HashableArray directly
    return x[..., np.asarray(p.permutation)]


@dispatch
def product(p: Permutation, q: Permutation):  # noqa: F811
    name = None if p._name is None and q._name is None else f"{p} @ {q}"
    return Permutation(p(np.asarray(q)), name)


@struct.dataclass
class PermutationGroup(FiniteGroup):
    r"""
    Collection of permutation operations acting on sequences of length :code:`degree`.

    Group elements need not all be of type :class:`netket.utils.group.Permutation`,
    only act as such on a sequence when called.

    The class can contain elements that are distinct as objects (e.g.,
    :code:`Identity()` and :code:`Translation((0,))`) but have identical action.
    Those can be removed by calling
    :meth:`~netket.utils.group.PermutationGroup.remove_duplicates`.
    """

    degree: int
    """Number of elements the permutations act on."""

    def __hash__(self):
        return super().__hash__()

    def _canonical(self, x: Element) -> Array:
        return x(np.arange(self.degree, dtype=int))

    def to_array(self) -> Array:
        r"""
        Convert the abstract group operations to an array of permutation indices.


        It returns a matrix where the `i`-th row contains the indices corresponding
        to the `i`-th group element. That is, :code:`self.to_array()[i, j]`
        is :math:`g_i^{-1}(j)`. Moreover,

        .. code::

            G = # this permutation group...
            V = np.arange(G.degree)
            assert np.all(G(V) == V[..., G.to_array()])

        Returns:
            A matrix that can be used to index arrays in the computational basis
            in order to obtain their permutations.
        """
        return self._canonical_array()

    def __array__(self, dtype=None) -> Array:
        return np.asarray(self.to_array(), dtype=dtype)

    def remove_duplicates(self, *, return_inverse=False) -> "PermutationGroup":
        r"""
        Returns a new :code:`PermutationGroup` with duplicate elements (that is,
        elements which represent identical permutations) removed.

        Args:
            return_inverse: If `True`, also return indices to reconstruct the original
                group from the result.

        Returns:
            The permutation group with duplicate elements removed. If
            :code:`return_inverse==True`, it also returns the indices needed to
            reconstruct the original group from the result.
        """
        if return_inverse:
            group, inverse = super().remove_duplicates(return_inverse=True)
        else:
            group = super().remove_duplicates(return_inverse=False)

        pgroup = PermutationGroup(group.elems, self.degree)

        if return_inverse:
            return pgroup, inverse
        else:
            return pgroup

    @struct.property_cached
    def inverse(self) -> Array:
        try:
            lookup = self._canonical_lookup()
            inverses = []
            # `np.argsort` on a 1D permutation list generates the inverse permutation
            # it acts along last axis by default, so can perform it on to_array()
            # `np.argsort` changes int32 to int64 on Windows,
            # and we need to change it back
            perms = self.to_array()
            invperms = np.argsort(perms).astype(perms.dtype)

            for invperm in invperms:
                inverses.append(lookup[HashableArray(invperm)])

            return np.asarray(inverses, dtype=int)
        except KeyError as err:
            raise RuntimeError(
                "PermutationGroup does not contain the inverse of all elements"
            ) from err

    @struct.property_cached
    def product_table(self) -> Array:
        perms = self.to_array()
        inverse = perms[self.inverse].squeeze()
        n_symm = len(perms)
        lookup = np.unique(np.column_stack((perms, np.arange(len(self)))), axis=0)

        product_table = np.zeros([n_symm, n_symm], dtype=int)
        for i, g_inv in enumerate(inverse):
            row_perms = perms[:, g_inv]
            row_perms = np.unique(
                np.column_stack((row_perms, np.arange(len(self)))), axis=0
            )
            # row_perms should be a permutation of perms, so identical after sorting
            if np.any(row_perms[:, :-1] != lookup[:, :-1]):
                raise RuntimeError(
                    "PermutationGroup is not closed under multiplication"
                )
            # match elements in row_perms to group indices
            product_table[i, row_perms[:, -1]] = lookup[:, -1]

        return product_table

    @property
    def shape(self) -> Shape:
        r"""
        Tuple :code:`(<# of group elements>, <degree>)`.

        Equivalent to :code:`self.to_array().shape`.
        """
        return (len(self), self.degree)

    def apply_to_id(self, x: Array):
        """Returns the image of indices `x` under all permutations"""
        return self.to_array()[self.inverse][:, x]


@dispatch
def product(A: PermutationGroup, B: PermutationGroup):  # noqa: F811
    if A.degree != B.degree:
        raise ValueError(
            "Incompatible groups (`PermutationGroup`s of different degree)"
        )
    return PermutationGroup(
        elems=[a @ b for a, b in itertools.product(A.elems, B.elems)], degree=A.degree
    )


@dispatch
def product(G: PermutationGroup, x: Array):  # noqa: F811
    return np.moveaxis(x[..., G.to_array()], -2, 0)



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax.numpy as jnp
import jax
import netket as nk

from netket.vqs import MCState, expect
from netket.stats import statistics as mpi_statistics
from netket import jax as nkjax

from .S2_operator import Renyi2EntanglementEntropy


@expect.dispatch
def Renyi2(vstate: MCState, op: Renyi2EntanglementEntropy, chunk_size: int | None):
    if op.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    samples = vstate.samples
    n_chains = samples.shape[0]
    n_samples = samples.shape[0] * samples.shape[1]

    if n_chains % 2 != 0 and not vstate.sampler.is_exact:
        raise ValueError("Use an even number of chains.")

    if n_chains == 1:
        if n_samples % 2 != 0:
            samples = samples[:, :-1]
        σ_η = samples[:, : (n_samples // 2)]
        σp_ηp = samples[:, (n_samples // 2) :]

    else:
        σ_η = samples[: (n_chains // 2)]
        σp_ηp = samples[(n_chains // 2) :]

    return Renyi2_sampling_MCState(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        σ_η,
        σp_ηp,
        op.partition,
        chunk_size=chunk_size,
    )


@partial(jax.jit, static_argnames=("afun", "chunk_size"))
def Renyi2_sampling_MCState(
    afun, params, model_state, σ_η, σp_ηp, partition, *, chunk_size
):
    n_chains = σ_η.shape[0]

    N = σ_η.shape[-1]

    σ_η = σ_η.reshape(-1, N)
    σp_ηp = σp_ηp.reshape(-1, N)

    n_samples = σ_η.shape[0]

    σ = σ_η[:, partition]
    σp = σp_ηp[:, partition]

    σ_ηp = jnp.copy(σp_ηp)
    σp_η = jnp.copy(σ_η)

    σ_ηp = σ_ηp.at[:, partition].set(σ)
    σp_η = σp_η.at[:, partition].set(σp)

    @partial(
        nkjax.apply_chunked, in_axes=(None, None, 0, 0, 0, 0), chunk_size=chunk_size
    )
    def kernel_fun(params, model_state, σ_ηp, σp_η, σ_η, σp_ηp):
        W = {"params": params, **model_state}

        return jnp.exp(afun(W, σ_ηp) + afun(W, σp_η) - afun(W, σ_η) - afun(W, σp_ηp))

    kernel_values = kernel_fun(params, model_state, σ_ηp, σp_η, σ_η, σp_ηp)

    Renyi2_stats = mpi_statistics(kernel_values.reshape((n_chains, -1)).T)

    # Propagation of errors from S_2 to -log_2(S_2)
    Renyi2_stats = Renyi2_stats.replace(
        variance=Renyi2_stats.variance / (Renyi2_stats.mean.real * jnp.log(2)) ** 2
    )

    Renyi2_stats = Renyi2_stats.replace(
        error_of_mean=jnp.sqrt(
            Renyi2_stats.variance / (n_samples * nk.utils.mpi.n_nodes)
        )
    )

    Renyi2_stats = Renyi2_stats.replace(mean=-jnp.log2(Renyi2_stats.mean).real)

    return Renyi2_stats



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .abstract_hilbert import AbstractHilbert

import numpy as np


class ContinuousHilbert(AbstractHilbert):
    """Abstract class for the Hilbert space of particles
    in continuous space.

    This class defines the common interface that
    can be used to interact with particles defined
    in continuous space.
    """

    def __init__(self, domain: tuple[float, ...], pbc: bool | tuple[bool, ...]):
        """
        Constructs new ``Particles`` given specifications of the continuous
        space they are defined in.

        This object returns an Hilber space.

        Args:
            domain: Tuple indicating the maximum of the continuous quantum number(s) in the configurations. Each entry
                in the tuple corresponds to a different physical dimension.
                If np.inf is used an infinite box is considered and `pbc=False` is mandatory (because what are PBC
                if there are no boundaries?). If a finite value is given, a minimum value of zero is assumed for the
                quantum number(s).
                A particle in a 3D box of size L would take `(L,L,L)`. A rotor model would take e.g. `(2pi,)`.
            pbc: Tuple or bool indicating whether to use periodic boundary conditions in a given physical dimension.
                If tuple it must have the same length as domain. If bool the same value is used for all the dimensions
                defined in domain.
        """
        self._extent = tuple(domain)
        self._pbc = tuple(pbc)
        if not len(self._extent) == len(self._pbc):
            raise ValueError(
                """`pbc` must be either a bool or a tuple indicating the periodicity of each spatial dimension."""
            )

        if np.any(np.logical_and(np.isinf(self._extent), pbc)):
            raise ValueError(
                "If you do have periodic boundary conditions in a given direction the maximum of the quantum number "
                "in that direction must be finite."
            )

        super().__init__()

    @property
    def extent(self) -> tuple[float, ...]:
        r"""Spatial extension in each spatial dimension"""
        return self._extent

    @property
    def pbc(self) -> tuple[bool, ...]:
        r"""Whether or not to use periodic boundary conditions for each spatial dimension"""
        return self._pbc



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from textwrap import dedent

import jax.numpy as jnp
import numpy as np

from netket.hilbert import HomogeneousHilbert

from netket.operator._abstract_observable import AbstractObservable


class Renyi2EntanglementEntropy(AbstractObservable):
    r"""
    Rényi2 bipartite entanglement entropy of a state :math:`| \Psi \rangle`
    between partitions A and B.

    """

    def __init__(
        self,
        hilbert: None,
        partition: jnp.array,
    ):
        r"""
        Constructs the operator computing the Rényi2 entanglement entropy of
        a state :math:`| \Psi \rangle` for a partition with partition A:

        .. math::

            S_2 = -\log_2 \text{Tr}_A [\rho^2]

        where :math:`\rho = | \Psi \rangle \langle \Psi |` is the density
        matrix of the system and :math:`\text{Tr}_A` indicates the partial
        trace over the partition A.

        The Monte Carlo estimator of S_2 [Hastings et al., PRL 104, 157201 (2010)] is:

        .. math::

            S_2 = - \log \langle \frac{\Psi(\sigma,\eta^{\prime}) \Psi(\sigma^{\prime},\eta)}{\Psi(\sigma,\eta) \Psi(\sigma^{\prime},\eta^{\prime})} \rangle

        where the mean is taken over the distribution
        :math:`\Pi(σ,η) \Pi(σ',η')`, :math:`\sigma \in A`,
        :math:`\eta \in \bar{A}` and
        :math:`\Pi(\sigma, \eta) = |\Psi(\sigma,\eta)|^2 / \langle \Psi | \Psi \rangle`.

        Args:
            hilbert: hilbert space of the system.
            partition: list of the indices identifying the degrees of
                freedom in one partition of the full system. All
                indices should be integers between 0 and hilbert.size

        Returns:
            Rényi2 operator for which computing the expected value.
        """

        # Homogeneos, not discrete... because we don't support
        # tensorhilbert and such. We could generalize easily by
        # setting psi(x not in hilbert) = 0
        if not isinstance(hilbert, HomogeneousHilbert):
            raise TypeError(
                dedent(
                    """
                    Entanglement Entropy estimation is only implemented for
                    Homogeneous discrete Hilbert spaces.

                    It can be easily generalised to continuous spaces, so if
                    you want this feature get in touch with us!"
                    """
                )
            )
        else:
            if hilbert.constrained:
                raise ValueError(
                    dedent(
                        """
                        Entanglement entropy estimation is not implemented
                        for constrained Hilbert spaces.

                        It can be generalised, so get in touch with us if you
                        need this feature.
                        """
                    )
                )

        super().__init__(hilbert)

        self._partition = np.array(list(set(partition)))

        if (
            np.where(self._partition < 0)[0].size > 0
            or np.where(self._partition > hilbert.size - 1)[0].size > 0
        ):
            raise ValueError(
                "Invalid partition: possible negative indices or indices outside the system size."
            )

    @property
    def partition(self):
        r"""
        list of indices for the degrees of freedom in the partition
        """
        return self._partition

    def __repr__(self):
        return f"Renyi2EntanglementEntropy(hilbert={self.hilbert}, partition={self.partition})"



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from netket.utils import struct


class Timeout(struct.Pytree, mutable=True):
    """A simple callback to stop NetKet after some time has passed.

    This callback monitors whether a driver has been training for more
    than a given timeout in order to hard stop training.
    """

    timeout: float
    """Number of seconds to wait before the training will be stopped."""

    # caches
    _init_time: float | None = struct.field(serialize=False)
    """
    Internal field storing the time at which the first iteration has been
    performed.
    """

    def __init__(self, timeout: float):
        """
        Stops the optimisation after a certain time itnerval.

        Args:
            timeout: number of seconds after which the optimisation will
                be stopped.
        """
        if not timeout > 0:
            raise ValueError("`timeout` must be larger than 0.")
        self.timeout = timeout

        self._init_time = None

    def __call__(self, step, log_data, driver):
        """
        A boolean function that determines whether or not to stop training.

        Args:
            step: An integer corresponding to the step (iteration or epoch) in training.
            log_data: A dictionary containing log data for training.
            driver: A NetKet variational driver.

        Returns:
            A boolean. If True, training continues, else, it does not.

        Note:
            This callback does not make use of `step`, `log_data` nor `driver`.
        """
        if self._init_time is None:
            self._init_time = time.time()

        if time.time() - self._init_time >= self.timeout:
            return False
        else:
            return True



import sys
import re

from ._common import exec_in_terminal

PLATFORM = sys.platform


def available_cpus():
    """
    Detects the number of logical CPUs subscriptable by this process.
    On Linux, this checks /proc/self/status for limits set by
    taskset, on other platforms taskset do not exist so simply uses
    multiprocessing.

    This should be a good estimate of how many cpu cores Jax/XLA sees.
    """
    if PLATFORM.startswith("linux"):
        try:
            m = re.search(
                r"(?m)^Cpus_allowed:\s*(.*)$", open("/proc/self/status").read()
            )
            if m:
                res = bin(int(m.group(1).replace(",", ""), 16)).count("1")
                if res > 0:
                    return res
        except OSError:
            pass
    else:
        import multiprocessing

        return multiprocessing.cpu_count()


## Taken from https://github.com/matthew-brett/x86cpu/blob/530f89f678aba501381e94be5325b9c91b878a32/x86cpu/tests/info_getters.py#L24
##


def cpu_info():
    if PLATFORM.startswith("darwin"):
        return get_sysctl_cpu()
    elif PLATFORM.startswith("linux"):
        return get_proc_cpuinfo()
    elif PLATFORM.startswith("win32"):
        return get_wmic_cpu()
    else:
        raise ValueError(f"Unsupported platform {PLATFORM}")


SYSCTL_KEY_TRANSLATIONS = {
    "model": "model_display",
    "family": "family_display",
    "extmodel": "extended_model",
    "extfamily": "extended_family",
}


SYSCTL_FLAG_TRANSLATIONS = {
    "sse4.1": "sse4_1",
    "sse4.2": "sse4_2",
}


def get_sysctl_cpu():
    sysctl_text = exec_in_terminal(["sysctl", "-a"])
    info = {}
    for line in sysctl_text.splitlines():
        if not line.startswith("machdep.cpu."):
            continue
        line = line.strip()[len("machdep.cpu.") :]
        key, value = line.split(": ", 1)
        key = SYSCTL_KEY_TRANSLATIONS.get(key, key)
        try:
            value = int(value)
        except ValueError:
            pass
        info[key] = value
    # features is absent in M1 macs
    info_features = info.get("features", "")
    flags = [flag.lower() for flag in info_features.split()]
    info["flags"] = [SYSCTL_FLAG_TRANSLATIONS.get(flag, flag) for flag in flags]
    info["unknown_flags"] = ["3dnow"]
    info["supports_avx"] = "hw.optional.avx1_0: 1\n" in sysctl_text
    info["supports_avx2"] = "hw.optionxwal.avx2_0: 1\n" in sysctl_text
    return info


PCPUINFO_KEY_TRANSLATIONS = {
    "vendor_id": "vendor",
    "model": "model_display",
    "family": "family_display",
    "model name": "brand",
    "cpu cores": "core_count",
    "cpu_cores": "core_count",
}


def get_proc_cpuinfo():
    with open("/proc/cpuinfo") as fobj:
        pci_lines = fobj.readlines()
    info = {}
    for line in pci_lines:
        line = line.strip()
        if line == "":  # End of first processor
            break
        key, value = line.split(":", 1)
        key, value = key.strip(), value.strip()
        key = PCPUINFO_KEY_TRANSLATIONS.get(key, key)
        try:
            value = int(value)
        except ValueError:
            pass
        info[key] = value
    info["flags"] = info["flags"].split()
    # cpuinfo records presence of Prescott New Instructions, Intel's code name
    # for SSE3.
    if "pni" in info["flags"]:
        info["flags"].append("sse3")
    info["unknown_flags"] = ["3dnow"]
    info["supports_avx"] = "avx" in info["flags"]
    info["supports_avx2"] = "avx2" in info["flags"]
    return info


WMIC_KEY_TRANSLATIONS = {
    "manufacturer": "vendor",
    "model": "model_display",
    "level": "family_display",
    "name": "brand",
    "numberofcores": "core_count",
}


def get_wmic_cpu():
    """Get CPU parameters using ``wmic`` Windows utility
    For a description of each CPU field, see:
    https://msdn.microsoft.com/en-us/library/aa394373(v=vs.85).aspx
    """
    wmic_text = exec_in_terminal(
        ["wmic", "cpu", "get", "/all", "/format:textvaluelist"]
    )
    info = {}
    for line in wmic_text.splitlines():
        line = line.strip()
        if line == "":
            continue
        key, value = line.split("=", 1)
        key = key.lower()
        key = WMIC_KEY_TRANSLATIONS.get(key, key)
        try:
            value = int(value)
        except ValueError:
            pass
        if key in info:  # Now we're looking at another processor
            break
        info[key] = value

    # Stepping sometimes the empty string in wmic output
    if "stepping" in info and info["stepping"] == "":
        info["stepping"] = None

    # Get extra information from kernel32
    from ctypes import windll, wintypes

    has_feature = windll.kernel32.IsProcessorFeaturePresent
    has_feature.argtypes = [wintypes.DWORD]
    info["flags"] = {
        "sse": has_feature(6),
        "sse2": has_feature(10),
        "sse3": has_feature(13),  # Not available on XP
        "mmx": has_feature(3),
        "3dnow": has_feature(7),
    }
    info["unknown_flags"] = ("ssse3", "sse4_1", "sse4_2")

    # TODO: `IsProcessorFeaturePresent` cannot detect AVX.
    # Maybe we can use another package like `cpuinfo` for this.
    info["supports_avx"] = "Unknown"
    info["supports_avx2"] = "Unknown"

    return info



# Copyright 2021 The NetKet Authors - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import jax
import jax.numpy as jnp

from netket.utils.types import Array
from netket.utils.struct import field
from .._utils import expand_dim
from ._tableau import TableauRKExplicit
from .._solver import (
    AbstractSolver,
    AbstractSolverState,
    append_docstring,
    args_adaptive_docstring,
    args_fixed_dt_docstring,
)


class RKExplicitSolver(AbstractSolver):
    r"""
    Class representing the Butcher tableau of an explicit Runge-Kutta method [1,2],
    which, given the ODE :math:`dy/dt = F(t, y)`, updates the solution as

    .. math::
        y_{t+dt} = y_t + \sum_l b_l k_l

    with the intermediate slopes

    .. math::
        k_l = F(t + c_l dt, y_t + \sum_{m < l} a_{lm} k_m).

    If :code:`self.is_adaptive`, the tableau also contains the coefficients :math:`b'_l`
    which can be used to estimate the local truncation error by the formula

    .. math::
        y_{\mathrm{err}} = \sum_l (b_l - b'_l) k_l.

    [1] https://en.wikipedia.org/w/index.php?title=Runge%E2%80%93Kutta_methods&oldid=1055669759
    [2] J. Stoer and R. Bulirsch, Introduction to Numerical Analysis, Springer NY (2002).
    """

    tableau: TableauRKExplicit = field(pytree_node=False)
    """The Butcher tableau containing all coefficients for solving the ODE."""

    def __init__(self, dt, tableau, adaptive=False, **kwargs):
        self.tableau = tableau
        if adaptive and not tableau.is_adaptive:
            raise AttributeError(f"Tableau of type {tableau} cannot be adaptve.")
        super().__init__(dt=dt, adaptive=adaptive, **kwargs)

    def __repr__(self) -> str:
        return "{}(tableau={}, dt={}, adaptive={}, integrator_parameters={})".format(
            "RKExplicitSolver",
            self.tableau,
            self.dt,
            self.adaptive,
            self.integrator_params,
        )

    @property
    def is_explicit(self):
        """Boolean indication whether the integrator is explicit."""
        return jnp.allclose(
            self.tableau.a, jnp.tril(self.tableau.a)
        )  # check if lower triangular

    @property
    def is_adaptive(self):
        """Boolean indication whether the integrator can be adaptive."""
        return self.tableau.is_adaptive

    @property
    def stages(self):
        """
        Number of stages (equal to the number of evaluations of the ode function)
        of the scheme.
        """
        return len(self.tableau.c)

    @property
    def error_order(self):
        """
        Returns the order of the embedded error estimate for a tableau
        supporting adaptive step size. Otherwise, None is returned.
        """
        if not self.is_adaptive:
            return None
        else:
            return self.tableau.order[1]

    def _compute_slopes(
        self,
        f: Callable,
        t: float,
        dt: float,
        y_t: Array,
    ):
        """Computes the intermediate slopes k_l."""
        times = t + self.tableau.c * dt

        # TODO: Use FSAL

        k = expand_dim(y_t, self.stages)
        for l in range(self.stages):
            dy_l = jax.tree_util.tree_map(
                lambda k: jnp.tensordot(
                    jnp.asarray(self.tableau.a[l], dtype=k.dtype), k, axes=1
                ),
                k,
            )
            y_l = jax.tree_util.tree_map(
                lambda y_t, dy_l: jnp.asarray(y_t + dt * dy_l, dtype=dy_l.dtype),
                y_t,
                dy_l,
            )
            k_l = f(times[l], y_l, stage=l)
            k = jax.tree_util.tree_map(lambda k, k_l: k.at[l].set(k_l), k, k_l)

        return k

    def step(
        self, f: Callable, dt: float, t: float, y_t: Array, state: AbstractSolverState
    ):
        """Perform one fixed-size RK step from `t` to `t + dt`."""
        k = self._compute_slopes(f, t, dt, y_t)

        b = self.tableau.b[0] if self.tableau.b.ndim == 2 else self.tableau.b
        y_tp1 = jax.tree_util.tree_map(
            lambda y_t, k: y_t
            + jnp.asarray(dt, dtype=y_t.dtype)
            * jnp.tensordot(jnp.asarray(b, dtype=k.dtype), k, axes=1),
            y_t,
            k,
        )

        return y_tp1, state

    def step_with_error(
        self, f: Callable, dt: float, t: float, y_t: Array, state: AbstractSolverState
    ):
        """
        Perform one fixed-size RK step from `t` to `t + dt` and additionally return the
        error vector provided by the adaptive solver.
        """
        if not self.is_adaptive:
            raise RuntimeError(f"{self} is not adaptive")

        k = self._compute_slopes(f, t, dt, y_t)

        y_tp1 = jax.tree_util.tree_map(
            lambda y_t, k: y_t
            + jnp.asarray(dt, dtype=y_t.dtype)
            * jnp.tensordot(jnp.asarray(self.tableau.b[0], dtype=k.dtype), k, axes=1),
            y_t,
            k,
        )
        db = self.tableau.b[0] - self.tableau.b[1]
        y_err = jax.tree_util.tree_map(
            lambda k: jnp.asarray(dt, dtype=k.dtype)
            * jnp.tensordot(jnp.asarray(db, dtype=k.dtype), k, axes=1),
            k,
        )

        return y_tp1, y_err, state


@append_docstring(args_fixed_dt_docstring)
def Euler(dt):
    r"""
    The canonical first-order forward Euler method. Fixed timestep only.

    """
    from . import _tableau as rkt

    return RKExplicitSolver(dt, tableau=rkt.bt_feuler)


@append_docstring(args_fixed_dt_docstring)
def Midpoint(dt):
    r"""
    The second order midpoint method. Fixed timestep only.

    """
    from . import _tableau as rkt

    return RKExplicitSolver(dt, tableau=rkt.bt_midpoint)


@append_docstring(args_fixed_dt_docstring)
def Heun(dt):
    r"""
    The second order Heun's method. Fixed timestep only.

    """
    from . import _tableau as rkt

    return RKExplicitSolver(dt, tableau=rkt.bt_heun)


@append_docstring(args_fixed_dt_docstring)
def RK4(dt):
    r"""
    The canonical Runge-Kutta Order 4 method. Fixed timestep only.

    """
    from . import _tableau as rkt

    return RKExplicitSolver(dt, tableau=rkt.bt_rk4)


@append_docstring(args_adaptive_docstring)
def RK12(dt, **kwargs):
    r"""
    The second order Heun's method. Uses embedded Euler method for adaptivity.
    Also known as Heun-Euler method.

    """
    from . import _tableau as rkt

    return RKExplicitSolver(dt, tableau=rkt.bt_rk12, **kwargs)


@append_docstring(args_adaptive_docstring)
def RK23(dt, **kwargs):
    r"""
    2nd order adaptive solver with 3rd order error control,
    using the Bogacki–Shampine coefficients

    """
    from . import _tableau as rkt

    return RKExplicitSolver(dt, tableau=rkt.bt_rk23, **kwargs)


@append_docstring(args_adaptive_docstring)
def RK45(dt, **kwargs):
    r"""
    Dormand-Prince's 5/4 Runge-Kutta method. (free 4th order interpolant).

    """
    from . import _tableau as rkt

    return RKExplicitSolver(dt, tableau=rkt.bt_rk4_dopri, **kwargs)
